import os
import sys
import re
import pandas as pd
import numpy as np

# import pysam
import multiprocessing as mp
from collections import defaultdict, OrderedDict, Counter
from more_itertools import grouper
from Bio import pairwise2
from Bio import SeqIO
from spacemake.util import rev_comp
from spacemake.util import fasta_chunks
import matplotlib

matplotlib.use("Agg")
import matplotlib.pyplot as plt

blocks = OrderedDict()
for fa_id, seq in fasta_chunks(open("blocks_combv2.fa")):
    blocks[fa_id] = seq
    blocks[fa_id + "_RC"] = rev_comp(seq)

block_names = list(blocks.keys())


def read_fq(fname):
    for name, seq, _, qual in grouper(open(fname), 4):
        yield name.rstrip()[1:], seq.rstrip(), qual.rstrip()


# def orient(seq, fw=blocks['P5'][10:25], rc=blocks['N70X_RC'][31:46]):
#     start = seq[:50]
#     f = fw in start
#     r = rc in start
#     if f and not r:
#         return seq
#     elif r and not f:
#         return rev_comp(seq)
#     else:
#         print(start, f, r)
#         return None


# sample = "sts_067_1"
sample = sys.argv[1]
# base_path = (
#     f"/data/rajewsky/projects/combbeads/three_segments/pacbio/sts_083_{sample}.fq"
# )
base_path = (
    f"/data/rajewsky/projects/combbeads/three_segments/pacbio/sts_067_{sample}.fq"
)

# base_path = f"/data/rajewsky/projects/slide_seq/projects/sts_067/raw_data/pacbio/{sample}.fastq"
# oriented = []
# names = []
# for name, seq, qual in read_fq(base_path):
#     s = orient(seq)
#     if s is not None:
#         oriented.append(s)

# lengths = np.array([len(s) for s in oriented])
# L_max = lengths.max()

aln_caches = {}
import logging

logging.basicConfig(level=logging.DEBUG)
import cache


# # Run this block to compute all alignments for later re-use
# o = sys.argv[2]
# aln = cache.CachedAlignments(sample, o, blocks[o])

# # for name, seq, qual in read_fq(base_path):
# for name, seq, qual in read_fq(base_path):
#     aln.align_or_load(name, seq)

# sys.exit(0)


def annotate():
    multi = cache.MultiAlignments(sample, blocks)
    for o in ["SMART", "P5", "OP1", "OP2", "OP3", "pT", "N70X"]:
        aln_caches[o] = cache.CachedAlignments(sample, o, blocks[o])

    print("qname\tL\toligo\tstart\tend\tscore")
    for name, seq, qual in read_fq(base_path):
        for start, end, label, score in multi.annotate(name, seq):
            # pass
            print(f"{name}\t{len(seq)}\t{label}\t{start}\t{end}\t{score}")


# # Run this block next, to create the annotation flat files
# annotate()
# sys.exit(0)


class AnnotatedSequences:
    def __init__(
        self,
        sample_name,
        blocks,
        path=base_path,
    ):
        self.sample_name = sample_name
        self.path = path
        self.logger = logging.getLogger("AnnotatedSequences")
        self.raw_sequences = self.load_raw_sequences()
        self.ann_db = self.load_annotation()
        self.signatures = self.extract_signatures_and_orient()

    def load_raw_sequences(self):
        self.logger.info(f"loading raw sequences for {self.sample_name}")
        raw_sequences = {}
        for qname, seq, qual in read_fq(self.path.format(sample_name=self.sample_name)):
            raw_sequences[qname] = seq
        return raw_sequences

    def load_annotation(self):
        self.logger.info(f"loading oligo annotation for {self.sample_name}")
        df = pd.read_csv(f"{self.sample_name}.annotation.tsv", sep="\t")
        qdata = {}
        for qname, grp in df.groupby("qname"):
            qdata[qname] = (
                tuple(grp["oligo"]),
                grp["start"].values,
                grp["end"].values,
                grp["score"].values,
            )
        return qdata

    def extract_signatures_and_orient(self):
        self.logger.info(f"extracting signatures for {self.sample_name}")
        signatures = {}
        for qname, (names, starts, ends, scores) in self.ann_db.items():
            seq = self.raw_sequences[qname]
            L = len(seq)
            if names[0].endswith("_RC"):
                # reverse complement the whole thing
                names = tuple([(n + "_RC").replace("_RC_RC", "") for n in names[::-1]])
                scores = scores[::-1]
                starts, ends = L - ends[::-1], L - starts[::-1]
                seq = rev_comp(seq)
                # store the RC'ed data
                self.raw_sequences[qname] = seq
                self.ann_db[qname] = (names, starts, ends, scores)

            signatures[qname] = names

        return signatures

    def filter_signatures(self, qsig):
        for qname, sig in self.signatures.items():
            if sig == qsig:
                yield qname

    def query_dimensions(self, qsig):
        qnames = []
        L = []
        starts = []
        ends = []
        scores = []

        for qname in self.filter_signatures(qsig):
            qnames.append(qname)
            L.append(len(self.raw_sequences[qname]))
            names, s, e, sc = self.ann_db[qname]
            starts.append(s)
            ends.append(e)
            scores.append(sc)

        return (
            np.array(qnames),
            np.array(L),
            np.array(starts),
            np.array(ends),
            np.array(scores),
        )

    def query_oligo_occurrences(self, oligo):
        qnames = []
        L = []
        starts = []
        ends = []
        scores = []

        for qname, sig in self.signatures.items():
            if oligo in sig:
                i = list(sig).index(oligo)

                qnames.append(qname)
                L.append(len(self.raw_sequences[qname]))
                names, s, e, sc = self.ann_db[qname]
                starts.append(s[i])
                ends.append(e[i])
                scores.append(sc[i])

        return (
            np.array(qnames),
            np.array(starts),
            np.array(ends),
            np.array(scores),
            np.array(L),
        )

    def fmt(self, qname):
        def render_label(label, start, end):
            return list(f"|{label.center(end - start - 2)}|")

        #
        seq = self.raw_sequences[qname]
        L = len(seq)
        buf = [" "] * L
        for row in self.ann_db[qname].itertuples():
            buf[row.start : row.end + 1] = render_label(
                row.oligo, row.start, row.end + 1
            )
        #
        return f"{seq}\n{''.join(buf)}"


sig_intact = ("P5", "SMART", "OP1", "OP2", "OP3", "pT", "N70X")
S = AnnotatedSequences(
    sys.argv[1],
    blocks,
    path=base_path,
)
qintact, qL, qstarts, qends, qsores = S.query_dimensions(sig_intact)


# estimate extension efficiencies
# n_reads = 0
stages = np.zeros(5, dtype=float)
pT = np.zeros(4)
for qname, sig in S.signatures.items():
    stages[0] += 1
    # print(sig)
    l = list(sig)
    if "SMART" in sig:
        l = l[sig.index("SMART") + 1 :]
        stages[1] += 1
        try:
            block = l.pop(0)
            if block == "pT":
                pT[0] += 1
            elif block == "OP1":
                stages[2] += 1
                block = l.pop(0)
                if block == "pT":
                    pT[1] += 1
                elif block == "OP2":
                    stages[3] += 1
                    block = l.pop(0)
                    if block == "pT":
                        pT[2] += 1
                    elif block == "OP3":
                        stages[4] += 1
                        block = l.pop(0)
                        if block == "pT":
                            pT[3] += 1
        except IndexError:
            # this just happens when l is empty
            pass

# print(stages)
# stages /= n_reads
freqs = stages[1:] / stages[0]
efficiencies = stages[2:] / stages[1:-1]
pT_rates = pT / stages[0]
# print(stages)
print("## estimated extension efficiencies")
print(
    "sample\tf_SMART\tf_OP1\tf_OP2\tf_OP3\tr_seg1\tr_seg2\tr_seg3\tpT_0\tpT_1\tpT_2\tpT_3"
)
print(
    f"{sample}\t"
    + "\t".join([f"{e:.3f}" for e in np.concatenate((freqs, efficiencies, pT_rates))])
)
sys.exit(0)


def analyze_recurrence(AS, oligo):
    # collect start positions of additional oligo occurrences in "intact" reads
    extra_dstart = []
    extra_dend = []
    extra_scores = []
    extra_L = []
    for qname, sig in AS.signatures.items():
        if sig[:5] == sig_intact[:5]:
            if oligo in sig[5:]:
                i = 5 + list(sig[5:]).index(oligo)
                names, starts, ends, scores = AS.ann_db[qname]
                L = len(AS.raw_sequences[qname])
                extra_dstart.append(starts[i])
                extra_dend.append(L - ends[i])
                extra_scores.append(scores[i])
                extra_L.append(L)
    #
    return (
        np.array(extra_dstart),
        np.array(extra_dend),
        np.array(extra_scores),
        np.array(extra_L),
    )


def align_stats(ann, oname, qmatches, pad=1):
    oseq = blocks[oname]
    n = 0
    matches = np.zeros(len(oseq), dtype=float)
    edits = defaultdict(lambda: defaultdict(int))
    for qn, s, e, score, l in zip(*qmatches):
        s = max(0, s - pad)
        e = min(l, e + pad)
        n += 1
        rseq = ann.raw_sequences[qn][s:e]
        res = cache.align(rseq, oseq, min_score=0)
        if not res:
            continue
        aln = res[0]
        aln_str = pairwise2.format_alignment(*aln, full_sequences=True).split("\n")
        i = 0
        j = 0
        # print(aln_str)
        for r, m, q in zip(aln_str[0], aln_str[1], aln_str[2]):
            # print(r, m, q, i, len(oseq))
            if q == " ":
                # outside of oligo match range
                continue
            if m == "|":
                # print(matches[i])
                matches[i] += 1
            else:
                edits[i][q + r] += 1
            if q != "-":
                i += 1
    return matches / n, edits


def select(
    qmatches,
    from_start_min=None,
    from_start_max=None,
    from_end_min=None,
    from_end_max=None,
    min_score=None,
):
    qnames, starts, ends, scores, L = qmatches
    mask = np.ones(len(qnames), dtype=np.bool)
    if from_start_min is not None:
        mask &= starts >= from_start_min
    if from_start_max is not None:
        mask &= starts <= from_start_max
    fe = L - ends
    if from_end_min is not None:
        mask &= fe >= from_end_min
    if from_end_max is not None:
        mask &= fe <= from_end_max
    if min_score is not None:
        mask &= scores >= min_score
    return (qnames[mask], starts[mask], ends[mask], scores[mask], L[mask])


def subsample(qmatches, n=100):
    data = np.array(qmatches, dtype=np.object)
    n_cols, N = data.shape
    I = np.random.choice(N, size=n, replace=False)
    return data[:, I]


def plot_edits_heatmap(ax, oname, edits):
    oseq = blocks[oname]
    N = len(oseq)
    counts = np.zeros((N, 5))
    x = np.arange(N)
    for i in x:
        for j, nn in enumerate("ACGT-"):
            if nn == oseq[i]:
                counts[i, j] = np.nan
            counts[i, j] = edits[i][oseq[i] + nn]
    im = ax.imshow(counts.T, interpolation="none", cmap="viridis")
    plt.colorbar(im, ax=ax, shrink=0.5)
    ax.set_xticks(x)
    ax.set_xticklabels(list(oseq))
    ax.set_yticks(range(5))
    ax.set_yticklabels(list("ACGT-"))


fig, (ax1, ax2) = plt.subplots(2, figsize=(5, 8), sharex=True)
# plot occurrences of OP1,OP2 regardless of other matches
exp1 = 1 + len(blocks["P5"] + blocks["SMART"]) + 9.5 + 2
exp2 = 1 + len(blocks["P5"] + blocks["SMART"] + blocks["OP1"]) + 9.5 + 8
smart = op1 = S.query_oligo_occurrences("SMART")
op1 = S.query_oligo_occurrences("OP1")
op2 = S.query_oligo_occurrences("OP2")
op3 = S.query_oligo_occurrences("OP3")
pT = S.query_oligo_occurrences("pT")
op1rc = S.query_oligo_occurrences("OP1_RC")
op2rc = S.query_oligo_occurrences("OP2_RC")
op3rc = S.query_oligo_occurrences("OP3_RC")

r2_bound = 43 + len(blocks["N70X"])


def prepare_analyses(min_score=None):
    analyses = [
        # # Opseq1
        # (
        #     "OP1",
        #     "correct",
        #     select(
        #         op1,
        #         from_start_min=exp1 - 5,
        #         from_start_max=exp1 + 5,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP1",
        #     "prox",
        #     select(
        #         op1,
        #         from_start_min=exp2 + len(blocks["OP2"]),
        #         from_start_max=exp2 + len(blocks["OP2"]) + 100,
        #         from_end_min=r2_bound,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP1",
        #     "internal",
        #     select(
        #         op1,
        #         from_start_min=exp2 + len(blocks["OP2"]) + 100,
        #         from_end_min=r2_bound,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP1",
        #     "read2",
        #     select(op1, from_end_max=r2_bound, min_score=min_score),
        # ),
        # # Opseq2
        # (
        #     "OP2",
        #     "correct",
        #     select(
        #         op2,
        #         from_start_min=exp2 - 5,
        #         from_start_max=exp2 + 5,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP2",
        #     "prox",
        #     select(
        #         op2,
        #         from_start_min=exp2 + len(blocks["OP2"]) + 5,
        #         from_start_max=exp2 + len(blocks["OP2"]) + 105,
        #         from_end_min=r2_bound,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP2",
        #     "internal",
        #     select(
        #         op2,
        #         from_start_min=exp2 + len(blocks["OP2"]) + 105,
        #         from_end_min=r2_bound,
        #         min_score=min_score,
        #     ),
        # ),
        # (
        #     "OP2",
        #     "read2",
        #     select(op2, from_end_max=r2_bound, min_score=min_score),
        # ),
        # SMART
        ("SMART", "anywhere", select(smart, min_score=min_score)),
        # Opseq1
        ("OP1", "anywhere", select(op1, min_score=min_score)),
        # Opseq2
        ("OP2", "anywhere", select(op2, min_score=min_score)),
        # Opseq3
        ("OP3", "anywhere", select(op3, min_score=min_score)),
        # polyT
        ("pT", "anywhere", select(pT, min_score=min_score)),
        # Opseq1 RC
        ("OP1_RC", "anywhere", select(op1rc, min_score=min_score)),
        # Opseq2 RC
        ("OP2_RC", "anywhere", select(op2rc, min_score=min_score)),
        # Opseq3 RC
        ("OP3_RC", "anywhere", select(op3rc, min_score=min_score)),
    ]
    return analyses


def plot_scores(analyses, suffix=""):
    ### match score statistics
    fig, ax = plt.subplots()
    for oname, region, matches in analyses:
        qnames, starts, ends, scores, L = matches
        x = sorted(scores)
        y = np.linspace(0, 1, len(scores))
        ax.plot(x, y, label=f"{oname} {region} n={len(scores)}")
        ax.set_ylabel("cumulative rel. freq")
        ax.set_xlabel("alignment score")
    #
    ax.legend()
    fig.suptitle(sample)
    # fig.tight_layout()
    fig.savefig(f"scores_{sample}{suffix}.pdf")
    plt.close("all")


def plot_starts(analyses, suffix=""):
    ### start position statistics
    # fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 12))
    fig, ax = plt.subplots()
    for oname, region, matches in analyses:
        qnames, starts, ends, scores, L = matches
        x = sorted(starts)
        y = np.linspace(0, 1, len(scores))
        ax.plot(x, y, label=f"{oname} {region} n={len(scores)}")
        ax.set_ylabel("cumulative rel. freq")
        ax.set_xlabel("start positions")
    #
    ax.legend()
    fig.suptitle(sample)
    # fig.tight_layout()
    fig.savefig(f"starts_{sample}{suffix}.pdf")
    plt.close("all")


def plot_readlengths(analyses, suffix=""):
    ### start position statistics
    # fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(2, 2, figsize=(12, 12))
    fig, ax = plt.subplots()
    for oname, region, matches in analyses:
        qnames, starts, ends, scores, L = matches
        x = sorted(L)
        y = np.linspace(0, 1, len(x))
        ax.plot(x, y, label=f"{oname} {region} n={len(scores)}")
        ax.set_ylabel("cumulative rel. freq")
        ax.set_xlabel("original read length")
    #
    ax.legend()
    fig.suptitle(sample)
    # fig.tight_layout()
    fig.savefig(f"readlengths_{sample}{suffix}.pdf")
    fig.savefig(f"readlengths_{sample}{suffix}.svg")
    plt.close("all")


def plot_mismatches(analyses, suffix=""):
    fig, ax = plt.subplots()
    fig2, axes = plt.subplots(len(analyses), figsize=(8, 12), sharex=True, sharey=True)
    if len(analyses) == 1:
        axes = [
            axes,
        ]
    #
    for (oname, region, matches), ax2 in zip(analyses, axes):
        if len(matches[0]) > 500:
            matches = subsample(matches, n=500)
        #
        qnames, starts, ends, scores, L = matches
        m, ed = align_stats(S, oname, matches)
        x = np.arange(len(m))
        ax.step(x, m, where="mid", lw=2, label=f"{oname} {region} n={len(qnames)}")
        #
        plot_edits_heatmap(ax2, oname, ed)
        ax2.title.set_text(f"{region} n={len(qnames)}")
        ax2.set_xlabel(f"{oname} sequence")
    #
    ax.set_xticks(x)
    ax.set_ylim(0.5, 1.0)
    ax.legend()
    ax.set_xticklabels(list(blocks[oname]))
    ax.set_ylabel("match frequency")
    ax.set_xlabel(f"{oname} oligo")
    fig.tight_layout()
    fig.savefig(f"match_{oname}_{sample}{suffix}.pdf")
    fig.savefig(f"match_{oname}_{sample}{suffix}.svg")
    fig2.suptitle(f"{sample} {oname} mismatch, deletion statistics by location")
    fig2.savefig(f"edits_{oname}_{sample}{suffix}.pdf")
    fig2.savefig(f"edits_{oname}_{sample}{suffix}.svg")
    plt.close("all")


def plot_startpos_hist(ax, analyses):
    ax.set_yscale("log")
    for oname, region, matches in analyses:
        qnames, starts, ends, scores, L = matches
        x = np.arange(starts.max() + 1)
        pos = np.bincount(starts)
        ax.step(
            x,
            pos,
            where="mid",
            label=f"{oname} {region} n={len(qnames)}",
            lw=1.5,
            alpha=0.75,
        )
    #
    ax.set_ylabel("match count")
    ax.set_xlabel("start position")
    ax.legend()


def plot_fromend_hist(ax, analyses):
    ax.set_yscale("log")
    for oname, region, matches in analyses:
        qnames, starts, ends, scores, L = matches
        fe = L - ends
        x = np.arange(fe.max() + 1)
        pos = np.bincount(fe)
        ax.step(
            x,
            pos,
            where="mid",
            label=f"{oname} {region} n={len(qnames)}",
            lw=1.5,
            alpha=0.75,
        )
    #
    ax.set_ylabel("match count")
    ax.set_xlabel("distance from distal end")
    ax.legend()


analyses = prepare_analyses()
# plot_scores(analyses)
# plot_mismatches(analyses[-2:-1])
# plot_mismatches(analyses[-1:])

# fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
#     2, 2, figsize=(8, 8), sharex=True, sharey=True
# )
# plot_startpos_hist(ax1, analyses[-4:-2])
# plot_fromend_hist(ax2, analyses[-4:-2])
# plot_startpos_hist(ax3, analyses[-2:])
# plot_fromend_hist(ax4, analyses[-2:])
# fig.suptitle(sample)
# fig.savefig(f"aln_pos_{sample}.pdf")
# plt.close("all")

# # using 67_1 as negative control (should not have OP2) 25 is a good cutoff
# analyses = prepare_analyses(min_score=25.0)
# plot_scores(analyses, suffix="_ms")
# plot_starts(analyses, suffix="_ms")
# plot_readlengths(analyses, suffix="_ms")
# plot_mismatches(analyses[:4], suffix="_ms")
# plot_mismatches(analyses[4:8], suffix="_ms")
# plot_mismatches(analyses[-2:-1], suffix="_ms")
# plot_mismatches(analyses[-1:], suffix="_ms")

# fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
#     2, 2, figsize=(8, 8), sharex=True, sharey=True
# )
# plot_startpos_hist(ax1, analyses[-4:-2])
# plot_fromend_hist(ax2, analyses[-4:-2])
# plot_startpos_hist(ax3, analyses[-2:])
# plot_fromend_hist(ax4, analyses[-2:])
# fig.suptitle(sample)
# fig.savefig(f"aln_pos_{sample}_ms.pdf")
# plt.close("all")
plot_scores(analyses)
plot_mismatches(analyses)
maxX = 1000

fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
    2, 2, figsize=(8, 8), sharex=True, sharey=True
)
plot_startpos_hist(ax1, analyses[:-3])
plot_fromend_hist(ax2, analyses[:-3])
plot_startpos_hist(ax3, analyses[-3:])
plot_fromend_hist(ax4, analyses[-3:])
fig.suptitle(sample)
ax1.set_xlim(0, maxX)
ax2.set_xlim(0, maxX)
ax3.set_xlim(0, maxX)
ax4.set_xlim(0, maxX)
fig.savefig(f"aln_pos_{sample}.pdf")
plt.close("all")

# using 67_1 as negative control (should not have OP2) 25 is a good cutoff
analyses = prepare_analyses(min_score=25.0)
plot_scores(analyses, suffix="_ms")
plot_starts(analyses, suffix="_ms")
plot_readlengths(analyses, suffix="_ms")
plot_mismatches(analyses, suffix="_ms")
# plot_mismatches(analyses[4:8], suffix="_ms")
# plot_mismatches(analyses[-2:-1], suffix="_ms")
# plot_mismatches(analyses[-1:], suffix="_ms")

fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(
    2, 2, figsize=(8, 12), sharex=True, sharey=True
)
plot_startpos_hist(ax1, analyses[:-3])
plot_fromend_hist(ax2, analyses[:-3])
plot_startpos_hist(ax3, analyses[-3:])
plot_fromend_hist(ax4, analyses[-3:])
ax1.set_xlim(0, maxX)
ax2.set_xlim(0, maxX)
ax3.set_xlim(0, maxX)
ax4.set_xlim(0, maxX)
fig.suptitle(sample)
fig.savefig(f"aln_pos_{sample}_ms.pdf")
plt.close("all")

### mismatch and deletion statistics for OP1
# op1_correct = select(
#     op1, from_start_min=exp1 - 5, from_start_max=exp1 + 5, min_score=ms_op1
# )
# m_corr, ed_corr = align_stats(S, "OP1", subsample(op1_correct, n=500))

# op1_read2 = select(op1, from_end_max=43 + len(blocks["N70X"]), min_score=ms_op1)
# m_r2, ed_r2 = align_stats(S, "OP1", np.array(op1_read2, dtype=np.object))

# op1_prox = select(
#     op1,
#     from_start_min=exp2 + len(blocks["OP2"]),
#     from_start_max=exp2 + len(blocks["OP2"]) + 100,
#     min_score=ms_op1,
# )
# m_prox, ed_prox = align_stats(S, "OP1", op1_prox)

# op1_internal = select(
#     op1,
#     from_start_min=exp2 + len(blocks["OP2"]) + 100,
#     from_end_min=43 + len(blocks["N70X"]),
#     min_score=ms_op1,
# )
# m_int, ed_int = align_stats(S, "OP1", op1_internal)


### occurrence statistics for OP1, OP2, OP1_RC and OP2_RC
fig, ((ax, ax3), (ax2, ax4)) = plt.subplots(2, 2)
ax.set_yscale("log")
ax2.set_yscale("log")
ax3.set_yscale("log")
ax4.set_yscale("log")

L_max = max(op1[4].max(), op2[4].max(), op3[4].max())
x = np.arange(L_max)
ax.axvline(exp1, label="expect OP1", lw=0.1, color="blue", ls="dashed")
ax.axvline(exp2, label="expect OP2", lw=0.1, color="orange", ls="dashed")
ax.step(
    x,
    np.bincount(op1[1], minlength=L_max),
    where="mid",
    label="OP1",
    alpha=0.5,
    color="blue",
)
ax.step(
    x,
    np.bincount(op2[1], minlength=L_max),
    where="mid",
    label="OP2",
    alpha=0.5,
    color="orange",
)
ax.step(
    x,
    np.bincount(op3[1], minlength=L_max),
    where="mid",
    label="OP3",
    alpha=0.5,
    color="red",
)
ax.set_ylabel("match count")
ax.set_xlabel("position from start of cDNA")
ax.legend()

L_max = max(op1rc[4].max(), op2rc[4].max())
x = np.arange(L_max)
ax2.axvline(exp1, label="expect OP1", lw=0.1, color="blue", ls="dashed")
ax2.axvline(exp2, label="expect OP2", lw=0.1, color="orange", ls="dashed")
ax2.step(
    x,
    np.bincount(op1rc[1], minlength=L_max),
    where="mid",
    label="OP1_RC",
    alpha=0.5,
    color="blue",
)
ax2.step(
    x,
    np.bincount(op2rc[1], minlength=L_max),
    where="mid",
    label="OP2_RC",
    alpha=0.5,
    color="orange",
)
ax2.step(
    x,
    np.bincount(op3rc[1], minlength=L_max),
    where="mid",
    label="OP3_RC",
    alpha=0.5,
    color="red",
)
ax2.set_ylabel("match count")
ax2.set_xlabel("position from start of cDNA")
ax2.legend()


L_max = max(op1[4].max(), op2[4].max(), op3[4].max())
x = np.arange(L_max)
ax3.step(
    x,
    np.bincount(op1[4] - op1[2], minlength=L_max),
    where="mid",
    label="OP1",
    alpha=0.5,
    color="blue",
)
ax3.step(
    x,
    np.bincount(op2[4] - op2[2], minlength=L_max),
    where="mid",
    label="OP2",
    alpha=0.5,
    color="orange",
)
ax3.step(
    x,
    np.bincount(op3[4] - op3[2], minlength=L_max),
    where="mid",
    label="OP3",
    alpha=0.5,
    color="red",
)

ax3.set_ylabel("match count")
ax3.set_xlabel("position from END of cDNA")
ax3.legend()

L_max = max(op1rc[4].max(), op2rc[4].max(), op3rc[4].max())
x = np.arange(L_max)
ax4.step(
    x,
    np.bincount(op1rc[4] - op1rc[2], minlength=L_max),
    where="mid",
    label="OP1_RC",
    alpha=0.5,
    color="blue",
)
ax4.step(
    x,
    np.bincount(op2rc[4] - op2rc[2], minlength=L_max),
    where="mid",
    label="OP2_RC",
    alpha=0.5,
    color="orange",
)
ax4.step(
    x,
    np.bincount(op3rc[4] - op3rc[2], minlength=L_max),
    where="mid",
    label="OP3_RC",
    alpha=0.5,
    color="red",
)

ax4.set_ylabel("match count")
ax4.set_xlabel("position from END of cDNA")
ax4.legend()

# ax2.hist(lengths, bins=bins, label="all", histtype='step', density=True)
# ax2.hist(lengths[m1 & m2], bins=bins, label="with OP1 & OP2", histtype='step', density=True)
# for func, name in analyses:
#     _, m1 = detect_matches(filter_func=func)
#     # ax2.hist(lengths[m1], bins=bins, label=f"with OP1 {name}", histtype='step', density=True)
#     ax2.step(sorted(lengths[m1]), np.linspace(0,1, m1.sum()), label=f"with OP1 {name} (n={m1.sum()})", where='mid')

# ax2.set_ylabel("cumulative rel. frequency")
# ax2.set_xlabel("PB read length")
# ax2.legend()
fig.tight_layout()
fig.savefig(f"aln_pos_match_{sample}.pdf")
plt.close("all")


rcOP1 = analyze_recurrence(S, "OP1")
rcOP1_RC = analyze_recurrence(S, "OP1_RC")
rcOP2_RC = analyze_recurrence(S, "OP2_RC")

fig, ((ax, ax2), (ax3, ax4)) = plt.subplots(2, 2)
ax.hist(rcOP1_RC[0], bins=100, histtype="step", label="OP1_RC", cumulative=True)
ax.hist(rcOP2_RC[0], bins=100, histtype="step", label="OP2_RC", cumulative=True)
ax.set_xlabel("distance from bead oligo start")
ax.set_ylabel("cumulative frequency")
ax.legend()

ax2.hist(rcOP1_RC[1], bins=100, histtype="step", label="OP1_RC", cumulative=True)
ax2.hist(rcOP2_RC[1], bins=100, histtype="step", label="OP2_RC", cumulative=True)
ax2.set_xlabel("distance from bead oligo end")
ax2.set_ylabel("cumulative frequency")
ax2.legend()

ax3.hist(rcOP1_RC[2], bins=100, histtype="step", label="OP1_RC", cumulative=True)
ax3.hist(rcOP2_RC[2], bins=100, histtype="step", label="OP2_RC", cumulative=True)
ax3.set_xlabel("oligo match score")
ax3.set_ylabel("cumulative frequency")
ax3.legend()

ax4.hist(rcOP1_RC[3], bins=100, histtype="step", label="OP1_RC", cumulative=True)
ax4.hist(rcOP2_RC[3], bins=100, histtype="step", label="OP2_RC", cumulative=True)
ax4.set_xlabel("length of PB-read with a match")
ax4.set_ylabel("cumulative frequency")
ax4.legend()

fig.suptitle(sample)
fig.tight_layout()
fig.savefig(f"recurrent_OPx_{sample}.pdf")
plt.close("all")


from collections import Counter

for sig, count in Counter(S.signatures.values()).most_common(10):
    print(f"{count}\t{sig}")

sys.exit(0)


def select_start(seq, x):
    return x < 110


def select_prox(seq, x):
    return (x > 110) and (x < 250)


def select_ds(seq, x):
    return x >= 250


def select_read2(seq, x):
    return x > len(seq) - (43 + len(blocks["N70X"]))


analyses = [
    (select_start, "start"),
    (select_prox, "prox"),
    (select_ds, "ds"),
    (select_read2, "read2"),
]


def detect_matches(pattern=blocks["OP1"][2:12], filter_func=lambda seq, x: True):
    pos = np.zeros(L_max)
    mask = np.zeros(len(oriented), dtype=np.bool)
    for i, seq in enumerate(oriented):
        for m in re.finditer(pattern, seq):
            x = m.start()
            if not filter_func(seq, x):
                continue
            pos[x] += 1
            mask[i] = True
    #
    return pos, mask


def detect_matches_reverse(
    pattern=blocks["OP1_RC"][-12:-2], filter_func=lambda seq, x: True
):
    pos = np.zeros(L_max)
    mask = np.zeros(len(oriented), dtype=np.bool)
    for i, seq in enumerate(oriented):
        seq = rev_comp(seq)
        for m in re.finditer(pattern, seq):
            x = m.start()
            if not filter_func(seq, x):
                continue
            pos[x] += 1
            mask[i] = True
    #
    return pos, mask


exp1 = 1 + len(blocks["P5"] + blocks["SMART"]) + 9.5 + 2
exp2 = 1 + len(blocks["P5"] + blocks["SMART"] + blocks["OP1"]) + 9.5 + 8

op1_pos, m1 = detect_matches()
op2_pos, m2 = detect_matches(pattern=blocks["OP2"][:10])

fig, (ax, ax2) = plt.subplots(2, sharex=True)
ax.set_yscale("log")
x = np.arange(L_max)
ax.axvline(exp1, label="expect OP1", lw=0.1, color="blue", ls="dashed")
ax.axvline(exp2, label="expect OP2", lw=0.1, color="orange", ls="dashed")
ax.step(x, op1_pos, where="mid", label="OP1", alpha=0.5, color="blue")
ax.step(x, op2_pos, where="mid", label="OP2", alpha=0.5, color="orange")
ax.set_ylabel("match count")
ax.set_xlabel("position in PB read")
ax.legend()

# subset_lengths1 = []
# subset_lengths2 = []
# for func, name in analyses:
#     _, m1 = detect_matches(filter_func=func)
#     _, m2 = detect_matches(filter_func=func, pattern=blocks['OP2'][:10])
#     subset_lengths1.append(lengths[m1])
#     subset_lengths2.append(lengths[m2])

bins = 100
# ax2.hist(lengths, bins=bins, label="all", histtype='step', density=True)
# ax2.hist(lengths[m1 & m2], bins=bins, label="with OP1 & OP2", histtype='step', density=True)
for func, name in analyses:
    _, m1 = detect_matches(filter_func=func)
    # ax2.hist(lengths[m1], bins=bins, label=f"with OP1 {name}", histtype='step', density=True)
    ax2.step(
        sorted(lengths[m1]),
        np.linspace(0, 1, m1.sum()),
        label=f"with OP1 {name} (n={m1.sum()})",
        where="mid",
    )

ax2.set_ylabel("cumulative rel. frequency")
ax2.set_xlabel("PB read length")
ax2.legend()
fig.tight_layout()
fig.savefig(f"pos_{sample}.pdf")
plt.close("all")


op1_distal, m1 = detect_matches_reverse()
op2_distal, m2 = detect_matches_reverse(pattern=blocks["OP2_RC"][-10:])

ill_start = len(blocks["N70X"])
ill_end = ill_start + 43

fig, ax = plt.subplots()
ax.set_yscale("log")
x = np.arange(L_max)
ax.axvline(ill_start, label="ill r2 start", lw=0.1, color="k", ls="dashed")
ax.axvline(ill_end, label="ill r2 end (43nt)", lw=0.1, color="k", ls="dashed")
ax.bar(x, op1_distal, label="OP1_RC", alpha=0.5, color="blue")
ax.bar(x, op2_distal, label="OP2_RC", alpha=0.5, color="orange")
ax.set_xlabel("position from end of PB read")
ax.set_ylabel("total count")
ax.legend()
fig.tight_layout()
fig.savefig(f"pos_distal_{sample}.pdf")
plt.close("all")


def nuc_count(mat, seq, pos, flank=100):
    idx = {
        "A": 0,
        "C": 1,
        "G": 2,
        "T": 3,
    }
    for x in range(max(0, pos - flank), min(pos + flank + 1, len(seq))):
        j = x - pos + flank
        nt = seq[x]
        if nt in idx:
            mat[idx[nt], j] += 1


def scan_sequence_matches(
    seqs, pos_filter=lambda seq, x: x > 250, pattern=blocks["OP1"][2:12]
):
    mat = np.ones((4, 201)) * 5  # 5 pseudo counts
    lengths = []
    for i, seq in enumerate(seqs):
        for m in re.finditer(pattern, seq):
            x = m.start()
            if pos_filter(seq, x):
                nuc_count(mat, seq, x)
                lengths.append(len(seq))
    #
    counts = mat.sum(axis=0)
    freq = mat / counts[np.newaxis, :]
    ent = 2 + np.where(freq > 0, freq * np.log2(freq), 0).sum(axis=0)
    return freq * ent[np.newaxis, :], np.array(lengths)


from RBPamp.affinitylogo import _draw_logo


def plot_seqlogo(ax, pfm, info=False, charwidth=1.0, **kwargs):
    if info:
        info_content = 2 - pfm.apply(lambda p: (-p * np.log2(p)).sum(), axis=1)
        matrix = pfm.mul(info_content, axis=0)
    else:
        matrix = pfm
    #
    _draw_logo(ax, matrix, charwidth, **kwargs)


for func, name in analyses:
    flank = 100
    mat1, L1 = scan_sequence_matches(
        np.array(oriented)[m1], pattern=blocks["OP1"][2:12], pos_filter=func
    )
    mat2, L2 = scan_sequence_matches(
        np.array(oriented)[m2], pattern=blocks["OP2"][:10], pos_filter=func
    )
    #
    fig, (ax1, ax2) = plt.subplots(2, figsize=(16, 4), sharey=True, sharex=True)
    fig.suptitle(f"{sample} {name} (n1={len(L1)} n2={len(L2)})")
    plot_seqlogo(ax1, pd.DataFrame(mat1.T, columns=list("ACGT")))
    plot_seqlogo(ax2, pd.DataFrame(mat2.T, columns=list("ACGT")))
    #
    x = np.arange(-flank, flank + 1, 20)
    ax2.set_xticks(x + flank)
    ax2.set_xticklabels(x)
    ax2.set_ylim(0, 2)
    ax2.set_xlim(0, 201)
    ax2.set_ylabel("bits")
    ax1.set_ylabel("bits")
    ax1.set_xlabel(f"pos rel to {name} OP1 ({blocks['OP1']})")
    ax2.set_xlabel(f"pos rel to {name} OP2 ({blocks['OP2']})")
    # ax2.set_ylim(0, 2)
    fig.tight_layout()
    fig.savefig(f"context_seqlogo_{sample}_{name}.pdf")
    plt.close()



import os
import yaml
import argparse
import re
import logging

from spacemake.errors import *
from spacemake.util import str2bool, assert_file, bool_in_str, message_aggregation
from spacemake.util import check_star_index_compatibility

logger_name = 'spacemake.config'

def get_puck_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False, add_help=False)
    parser.add_argument('--name',
        help='name of the puck', type=str, required=True)
    parser.add_argument('--width_um',
        type=float, required=required,
        help='width of the puck in microns')
    parser.add_argument('--spot_diameter_um',
        type=float, required=required,
        help='diameter of the spots in this puck, in microns')
    parser.add_argument('--barcodes',
        type=str, required=False,
        help='path to barcode file. of not provided the --puck_barcode_file variable' +
            ' of `spacemake projects add_sample` has to be set')

    return parser

def get_run_mode_parser(required=True):
    parser = argparse.ArgumentParser(
            allow_abbrev=False, formatter_class = argparse.RawTextHelpFormatter,
            description='add/update run_mode parent parser',
            add_help = False)
    parser.add_argument('--name', type=str,
            help='name of the run_mode to be added',
            required = True)
    parser.add_argument('--parent_run_mode', type=str,
            help='Name of the parent run_mode. All run_modes will fall back to \'default\'')
    parser.add_argument('--umi_cutoff', type=int, nargs='+',
            help='umi_cutoff for this run_mode.' +\
                    'the automated analysis will be run with these cutoffs')
    parser.add_argument('--n_beads',
            type=int,
            help='number of expected beads for this run_mode')
    parser.add_argument(
            '--clean_dge',
            required=False,
            choices=bool_in_str,
            type=str,
            help='if True, the DGE will be cleaned of barcodes which overlap with primers')
    parser.add_argument(
            '--detect_tissue',
            required=False,
            choices=bool_in_str,
            type=str,
            help='By default only beads having at least umi_cutoff UMI counts are analysed '+\
                    'during the automated analysis, all other beads are filtered out. If this ' +\
                    'parameter is set, contiguous islands within umi_cutoff passing beads will '+\
                    'also be included in the analysis')
    parser.add_argument(
            '--polyA_adapter_trimming',
            required=False,
            choices=bool_in_str,
            type=str,
            help='if set, reads will have polyA stretches and adapter sequence overlaps trimmed '+\
                    'BEFORE mapping.')
    parser.add_argument(
        '--count_intronic_reads',
        required=False,
        choices=bool_in_str,
        type=str,
        help='if set, INTRONIC reads will also be countsed (apart from UTR and CDS)')
    parser.add_argument('--count_mm_reads', required=False,
        choices=bool_in_str, type=str,
        help='if True, multi-mappers will also be counted. For every '+
            'multimapper only reads which have one unique read mapped' +
            'to a CDS or UTR region will be counted')
    
    parser.add_argument('--mesh_data', required=False,
        choices=bool_in_str, type=str,
        help ='if True, this data will be \'mehsed\': a hexagonal structured '+
            'meshgrid will be created, where each new spot will have diameter'+
            ' of --mesh_spot_diameter_um micron diameter and the spots will ' +
            'be spaced --mesh_spot_distance_um microns apart')
    parser.add_argument('--mesh_type', required=False,
        choices=['circle', 'hexagon'], type = str,
        help = 'circle: circles with diameter of --mesh_spot_diameter_um will be placed' +
                ' in a hex grid, where he distance between any two circles will be ' +
                '--mesh_spot_distance_um\n' +
                'hexagon: a mesh of touching hexagons will be created, with their ' +
                'centers being --mesh_spot_distance_um apart. This will cover all ' +
                'the data without any holes')
    parser.add_argument('--mesh_spot_diameter_um',
        type=float, required=False,
        help='diameter of mesh spot, in microns. to create a visium-style '+
            'mesh, use 55um')
    parser.add_argument('--mesh_spot_distance_um',
        type=float, required=False,
        help='distance between mesh spots in um. to create a visium-style '+
            'mesh use 100um')

    return parser

def get_species_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False, add_help=False)
    parser.add_argument('--name',
        help = 'name of the species to be added',
        type = str, required=True)
    parser.add_argument('--genome',
        help = 'path to the genome (.fa) file for the species to be added',
        type = str, required=required)
    parser.add_argument('--annotation',
        help = 'path to the annotation (.gtf) file for the species to be added',
        type = str, required=required)
    parser.add_argument('--rRNA_genome',
        help = 'path to the ribosomal-RNA genome (.fa) file for the species to be added',
        default=None,
        type = str)
    parser.add_argument('--STAR_index_dir',
        help = 'path to STAR index directory',
        type = str,
        required = False)

    return parser

def get_barcode_flavor_parser(required=True):
    parser = argparse.ArgumentParser(
        allow_abbrev=False,
        description='add/update barcode_flavor',
        add_help = False)
    parser.add_argument('--name',
        help = 'name of the barcode flavor', type = str,
        required = True)
    parser.add_argument('--umi',
        help = 'structure of UMI, using python\'s list syntax. Example: to set UMI to ' +
        '13-20 NT of Read1, use --umi r1[12:20]. It is also possible to use the first 8nt of '+
        'Read2 as UMI: --umi r2[0:8]',
        type=str, required=required)
    parser.add_argument('--cell_barcode',
        help = 'structure of CELL BARCODE, using python\'s list syntax. Example: to set'+
            ' the cell_barcode to 1-12 nt of Read1, use --cell_barcode r1[0:12]. It is also possible '+
            ' to reverse the CELL BARCODE, for instance with r1[0:12][::-1] (reversing the first 12nt of' +
            ' Read1, and assigning them as CELL BARCODE).',
        type=str, required=required)

    return parser

def get_variable_action_subparsers(config, parent_parser, variable):
    if variable == 'barcode_flavors':
        variable_singular = variable[:-1]
        variable_add_update_parser = get_barcode_flavor_parser
    elif variable == 'pucks':
        variable_singular = variable[:-1]
        variable_add_update_parser = get_puck_parser
    elif variable == 'run_modes':
        variable_singular = variable[:-1]
        variable_add_update_parser = get_run_mode_parser
    elif variable == 'species':
        variable_singular = variable
        variable_add_update_parser = get_species_parser
    

    command_help = {
        'list': f'list {variable} and their settings',
        'delete': f'delete {variable_singular}',
        'add': f'add a new {variable_singular}',
        'update': f'update an existing barcode_flavor'
    }

    # list command
    list_parser = parent_parser.add_parser(f'list_{variable}',
        description = command_help['list'],
        help = command_help['list'])
    list_parser.set_defaults(
        func=lambda args: list_variables_cmdline(config, args),
        variable=variable)

    func=lambda args: add_update_delete_variable_cmdline(config, args)

    # delete command
    delete_parser = parent_parser.add_parser(f'delete_{variable_singular}',
        description = command_help['delete'],
        help = command_help['delete'])
    delete_parser.add_argument('--name',
        help = f'name of the {variable_singular} to be deleted',
        type=str, required=True)
    delete_parser.set_defaults(
        func=func, action='delete',
            variable=variable)

    # add command
    add_parser = parent_parser.add_parser(f'add_{variable_singular}',
        parents=[variable_add_update_parser()],
        description = command_help['add'],
        help=command_help['add'])
    add_parser.set_defaults(
        func=func, action='add',
            variable=variable)

    # update command
    update_parser = parent_parser.add_parser(f'update_{variable_singular}',
        parents=[variable_add_update_parser(False)],
        description = command_help['update'],
        help = command_help['update'])
    update_parser.set_defaults(
        func=func, action='update',
            variable=variable)

def setup_config_parser(config, subparsers):
    parser_config = subparsers.add_parser('config', help = 'configure spacemake')
    parser_config_subparsers = parser_config.add_subparsers(help = 'config sub-command help')

    for variable in ConfigFile.main_variables_pl2sg.keys():
        get_variable_action_subparsers(
            config,
            parser_config_subparsers,
            variable)

    return parser_config

@message_aggregation(logger_name)
def add_update_delete_variable_cmdline(config, args):
    # set the name and delete from dictionary
    name = args['name']
    variable = args['variable']
    action = args['action']

    config.assert_main_variable(variable)

    # remove the args from the dict
    del args['action']
    del args['variable']
    del args['name']

    if action == 'add':
        func = config.add_variable
    elif action == 'update':
        func = config.update_variable
    elif action == 'delete':
        func = config.delete_variable

    var_variables = func(variable, name, **args)

    # print and dump config file
    config.logger.info(yaml.dump(var_variables, sort_keys=False))
    config.dump()

@message_aggregation(logger_name)
def list_variables_cmdline(config, args):
    variable = args['variable']
    del args['variable']

    config.logger.info(f'Listing {variable}')
    config.logger.info(yaml.dump(config.variables[variable]))

class ConfigMainVariable:
    def __init__(self, name, **kwargs):
        self.name = name
        self.variables = {}

        # loop over kwargs
        for variable_key, variable in kwargs.items():
            if variable_key not in self.variable_types:
                raise UnrecognisedConfigVariable(f'{variable_key}',
                    list(self.variable_types.keys()))
            elif self.variable_types[variable_key] == 'int_list':
                self.variables[variable_key] = [
                    int(x) for x in kwargs[variable_key]
                ]
            else:
                new_type = self.variable_types[variable_key]
                self.variables[variable_key] = new_type(kwargs[variable_key])

    def __str__(self):
        class_name = self.__class__.__name__
        return f'{class_name}: {self.name}. variables: {self.variables}'

    def update(self, other):
        self.variables.update(other.variables)


class RunMode(ConfigMainVariable):
    variable_types = {
        'parent_run_mode': str,
        'n_beads': int,
        'umi_cutoff': 'int_list',
        'clean_dge': bool,
        'detect_tissue': bool,
        'polyA_adapter_trimming': bool,
        'count_mm_reads': bool,
        'count_intronic_reads': bool,
        'mesh_data': bool,
        'mesh_type': str,
        'mesh_spot_diameter_um': int,
        'mesh_spot_distance_um': int
    }

    def has_parent(self):
        if 'parent_run_mode' in self.variables.keys():
            return True
        else:
            return False

    @property
    def parent_name(self):
        if self.has_parent():
            return self.variables['parent_run_mode']
        else:
            return None

class Puck(ConfigMainVariable):
    variable_types = {
        'barcodes': str,
        'spot_diameter_um': float,
        'width_um': int
    }

    @property
    def has_barcodes(self):
        return ('barcodes' in self.variables
                and self.variables['barcodes']
                and self.variables['barcodes'] != 'None')

class ConfigFile:
    initial_config_path = os.path.join(os.path.dirname(__file__),
        'data/config/config.yaml') 

    main_variables_pl2sg = {
        'pucks':'puck',
        'barcode_flavors':'barcode_flavor',
        'run_modes': 'run_mode',
        'species': 'species'
    }

    main_variables_sg2pl = {value: key for key, value in main_variables_pl2sg.items()}

    main_variable_sg2type = {
        'puck': str,
        'barcode_flavor': str,
        'run_mode': 'str_list',
        'species': str
    }

    def __init__(self):
        self.variables = {
            'root_dir': '',
            'temp_dir': '/tmp',
            'species': {},
            'barcode_flavors': {},
            'run_modes': {},
            'pucks': {},
        }
        self.file_path = 'config.yaml'
        self.logger = logging.getLogger(logger_name)

    @classmethod
    def from_yaml(cls, file_path):
        cf = cls()

        config_yaml_variables = None
        with open(file_path, 'r') as f:
            config_yaml_variables = yaml.load(f,
                    Loader=yaml.FullLoader)

        if config_yaml_variables is not None:
            cf.variables = config_yaml_variables

        cf.file_path = file_path

        if file_path != cf.initial_config_path:
            initial_config = ConfigFile.from_yaml(cf.initial_config_path)

            # correct variables to ensure backward compatibility
            cf.correct() 
            
            # check which variables do not exist, if they dont, 
            # copy them from initial config
            for main_variable in cf.main_variables_pl2sg:
                # update new main_variables
                if main_variable not in cf.variables:
                    cf.variables[main_variable] = initial_config.variables[main_variable]

            # deduce variables which have 'default' value. this is to ensure spacemake
            # always runs w/o errors downstream: ie when barcode flavor, run_mode or puck
            # is set to default
            cf.vars_with_default = [key for key, value in initial_config.variables.items()
                if 'default' in value]

            for var_with_default in cf.vars_with_default:
                default_val = initial_config.variables[var_with_default]['default']
                if 'default' not in cf.variables[var_with_default]:
                    cf.variables[var_with_default]['default'] = default_val
                else:
                    # update default run mode with missing values
                    default_val.update(cf.variables[var_with_default]['default'])
                    cf.variables[var_with_default]['default'] = default_val

        return cf

    def assert_main_variable(self, variable):
        if variable not in self.main_variables_pl2sg.keys():
            raise UnrecognisedConfigVariable(variable,
                list(self.main_variables_pl2sg.keys()))

    def correct(self):
        # ensures backward compatibility
        if 'pucks' not in self.variables:
            self.variables['pucks'] = {}
            if 'puck_data' in self.variables:
                if 'pucks' in self.variables['puck_data']:
                    self.variables['pucks'] = self.variables['puck_data']['pucks']
                    del self.variables['puck_data']['pucks']
        
        if 'barcode_flavors' not in self.variables:
            self.variables['barcode_flavors'] = {}
            if 'knowledge' in self.variables:
                if 'barcode_flavor' in self.variables['knowledge']:
                    self.variables['barcode_flavors'] = self.variables['knowledge']['barcode_flavor']
        
        if 'species' not in self.variables:
            # get all the species and save them in the right place
            # if species is empty, create a species dictionary
            self.variables['species'] = {}

            if 'knowledge' in self.variables:
                # extract all annotation info, if exists
                for species in self.variables['knowledge'].get('annotations', {}):
                    if species not in self.variables['species']:
                        self.variables['species'][species] = {}

                    self.variables['species'][species]['annotation'] = \
                        self.variables['knowledge']['annotations'][species]

                for species in self.variables['knowledge'].get('genomes', {}):
                    if species not in self.variables['species']:
                        self.variables['species'][species] = {}

                    self.variables['species'][species]['genome'] = \
                        self.variables['knowledge']['genomes'][species]

                for species in self.variables['knowledge'].get('rRNA_genomes', {}):
                    if species not in self.variables['species']:
                        self.variables['species'][species] = {}

                    self.variables['species'][species]['rRNA_genome'] = \
                        self.variables['knowledge']['rRNA_genomes'][species]

        if 'knowledge' in self.variables:
            del self.variables['knowledge']
        
        # correct run modes
        for run_mode_name, run_mode_variables in self.variables['run_modes'].items():
            variables = run_mode_variables.copy()
            for var in run_mode_variables:
                if not var in RunMode.variable_types:
                    del variables[var]

            self.variables['run_modes'][run_mode_name] = variables

    def dump(self):
        with open(self.file_path, 'w') as fo:
            fo.write(yaml.dump(self.variables))

    def set_file_path(self, file_path):
        self.file_path = file_path

    @property
    def puck_data(self):
        return self.variables['puck_data']

    def variable_exists(self, variable_name, variable_key):
        return variable_key in self.variables[variable_name]

    def assert_variable(self, variable_name, variable_key):
        self.assert_main_variable(variable_name)
        if not isinstance(variable_key, list):
            variable_key = [variable_key]

        for key in variable_key:
            if not self.variable_exists(variable_name, key):
                variable_name_sg = self.main_variables_pl2sg[variable_name]
                raise ConfigVariableNotFoundError(variable_name_sg,
                        key)

    def delete_variable(self, variable_name, variable_key):
        self.assert_variable(variable_name, variable_key)

        if variable_name in self.vars_with_default and variable_key == 'default':
            raise EmptyConfigVariableError(variable_name)

        variable_data = self.variables[variable_name][variable_key]

        del self.variables[variable_name][variable_key]

        return variable_data

    def process_run_mode_args(self, **kwargs):
        # typeset boolean values of run_mode
        default_run_mode = self.get_variable('run_modes', 'default')

        for key, value in default_run_mode.items():
            if isinstance(value, bool) and key in kwargs.keys():
                kwargs[key] = str2bool(kwargs[key])

        return kwargs

    def process_barcode_flavor_args(self, cell_barcode=None, umi=None):
        bam_tags = 'CR:{cell},CB:{cell},MI:{UMI},RG:{assigned}'

        # r(1|2) and then string slice
        to_match = r'r(1|2)(\[((?=-)-\d+|\d)*\:((?=-)-\d+|\d*)(\:((?=-)-\d+|\d*))*\])+$'

        if umi is not None and re.match(to_match, umi) is None:
            raise InvalidBarcodeStructure('umi', to_match)

        if cell_barcode is not None and re.match(to_match, cell_barcode) is None:
            raise InvalidBarcodeStructure('umi', to_match)

        barcode_flavor = {
                'bam_tags': bam_tags}

        if umi is not None:
            barcode_flavor['UMI'] = umi

        if cell_barcode is not None:
            barcode_flavor['cell'] = cell_barcode

        return barcode_flavor

    def process_species_args(
        self,
        genome=None,
        annotation=None,
        rRNA_genome = None,
        STAR_index_dir = None,
    ):
        assert_file(genome, default_value=None, extension = '.fa')
        assert_file(annotation, default_value=None, extension = '.gtf')
        assert_file(rRNA_genome, default_value = None, extension = '.fa')

        species = {}

        if genome is not None:
            species['genome'] = genome
        
        if annotation is not None:
            species['annotation'] = annotation

        if rRNA_genome is not None:
            species['rRNA_genome'] = rRNA_genome

        if STAR_index_dir is not None:
            check_star_index_compatibility(STAR_index_dir)
            species['STAR_index_dir'] = STAR_index_dir

        return species

    def process_puck_args(self, width_um=None, spot_diameter_um=None, barcodes = None):
        assert_file(barcodes, default_value = None, extension = 'all')
        
        puck = {}

        if width_um is not None:
            puck['width_um'] = float(width_um),

        if spot_diameter_um is not None:
            puck['spot_diameter_um'] = float(spot_diameter_um)

        if barcodes is not None:
            puck['barcodes'] = barcodes

        return puck

    def process_variable_args(self, variable, **kwargs):
        if variable == 'barcode_flavors':
            return self.process_barcode_flavor_args(**kwargs)
        elif variable == 'run_modes':
            return self.process_run_mode_args(**kwargs)
        elif variable == 'pucks':
            return self.process_puck_args(**kwargs)
        elif variable == 'species':
            return self.process_species_args(**kwargs)
        else:
            ValueError(f'Invalid variable: {variable}')
        
    def add_variable(self, variable, name, **kwargs):
        if not self.variable_exists(variable, name):
            values = self.process_variable_args(variable, **kwargs)
            self.variables[variable][name] = values
        else:
            if variable in ['run_modes', 'pucks', 'barcode_flavors']:
                # drop the last s
                variable = variable[:-1]

            raise DuplicateConfigVariableError(variable, name)

        return kwargs

    def update_variable(self, variable, name, **kwargs):
        if self.variable_exists(variable, name):
            values = self.process_variable_args(variable, **kwargs)
            self.variables[variable][name].update(values)

            variable_data = self.variables[variable][name]
        else:
            if variable in ['run_modes', 'pucks', 'barcode_flavors']:
                # drop the last s
                variable = variable[:-1]

            raise ConfigVariableNotFoundError(variable, name)

        return variable_data

    def get_variable(self, variable, name):
        if not self.variable_exists(variable, name):
            raise ConfigVariableNotFoundError(variable, name)
        else:
            return self.variables[variable][name]

    def get_run_mode(self, name):
        # first load the default values
        rm = RunMode(name,
            **self.get_variable('run_modes', name))

        default_rm = RunMode('default',
            **self.get_variable('run_modes', 'default'))

        if rm.has_parent():
            parent_rm = self.get_run_mode(rm.parent_name)
            parent_rm.update(rm)
            rm = parent_rm

        # update with self
        default_rm.update(rm)
        rm = default_rm

        return rm

    def get_puck(self, name, return_empty=False):
        try:
            return Puck(name, **self.get_variable('pucks', name))
        except ConfigVariableNotFoundError as e:
            if not return_empty:
                raise
            else:
                return Puck(name)






#!/usr/bin/env python3
__version__ = "0.9"
__author__ = ["Marvin Jens"]
__license__ = "GPL"
__email__ = ["marvin.jens@mdc-berlin.de"]

import argparse
import logging
import time
import os
import sys
from numpy.core.fromnumeric import argsort
import pandas as pd
import numpy as np
import pysam
import multiprocessing as mp
from collections import defaultdict
from Bio import pairwise2
from Bio import SeqIO

from spacemake.parallel import (
    put_or_abort,
    queue_iter,
    join_with_empty_queues,
    chunkify,
    ExceptionLogging,
)
from spacemake.util import read_fq

NO_CALL = "NNNNNNNN"

# TODO:
# * add support for three segments
# * turn Output into a context manager
# * load params from YAML (code from opseq)
# * save params to YAML in run-folder to document
# * get rid of ugly global NO_CALL
# * refactor into multiple modules?


def read_source(args):
    if args.read2:
        for (id1, seq1, qual1), (id2, seq2, qual2) in zip(
            read_fq(args.read1), read_fq(args.read2)
        ):
            # assert id1.split()[0] == id2.split()[0]
            yield id1, seq1, id2, seq2, qual2
    else:
        for id1, seq1, qual1 in read_fq(args.read1):
            yield id1, seq1, id1, "READ2 IS NOT AVAILABLE", "READ2 IS NOT AVAILABLE"


def hamming(seqA, seqB, costs, match=2):
    score = 0
    for a, b, c in zip(seqA, seqB, costs):
        if a == b:
            score += match
        else:
            if a == "A" and c == 0:
                # we are looking at the BC2 start position
                # This tweak gives almost ~1% point more matches 72.58% -> 73.43%
                score += 1
            score -= c
    return score


class BarcodeMatcher:
    def __init__(self, fname, length_specific=True, place="left"):
        self.logger = logging.getLogger("BarcodeMatcher")
        self.length_specific = length_specific
        self.names, self.seqs = self.load_targets(fname)
        self.slen = np.array([len(s) for s in self.seqs])
        if len(self.names) == 0:
            self.logger.warning(f"no references loaded! Disabling matching for {fname}")
            self.align = self.align_na
        else:
            self.lmin = self.slen.min()
            self.lmax = self.slen.max()
            self.place = place
            self.costs = {}
            self.slen_masks = {}
            for l in range(self.lmin, self.lmax + 1):
                self.slen_masks[l] = (self.slen == l).nonzero()[0]
                cost = np.ones(l)
                if place == "left":
                    cost[-8:] = 2
                    cost[-4:] = 3  # the diagnostic 4-mer

                if place == "right":
                    cost[0] = 0  # the error-prone first base
                    cost[1:8] = 2
                    cost[4:8] = 3  # the diagnostic 4-mer

                self.costs[l] = cost

            self.logger.debug(
                f"initialized from {len(self.names)} sequences from lmin={self.lmin} to lmax={self.lmax}"
            )

    def load_targets(self, fname):
        self.logger.debug(f"loading target sequences from '{fname}'")
        names = []
        seqs = []
        if fname:
            for rec in SeqIO.parse(fname, "fasta"):
                names.append(rec.id.split()[0])
                seqs.append(str(rec.seq).upper())

        return np.array(names), np.array(seqs)

    def align_na(self, query, debug=False):
        return (
            [NO_CALL],
            [NO_CALL],
            [-1],
        )

    def align(self, query, debug=False):
        results = []
        scores = []
        # select set of barcode sequences to align with
        if len(query) < self.lmin:
            return (
                [NO_CALL],
                [NO_CALL],
                [-1],
            )

        if self.length_specific:
            lq = min(len(query), self.lmax)
            lmask = self.slen_masks[lq]
            seqs_sel = self.seqs[lmask]
            names_sel = self.names[lmask]
        else:
            seqs_sel = self.seqs
            names_sel = self.names

        costs = self.costs[lq]
        # perform the alignments and keep results
        for seq in seqs_sel:
            res = hamming(query, seq, costs)
            # print(query, seq, res)
            results.append(res)
            scores.append(res)

        # identify best alignments and return tied, best barcode matches
        scores = np.array(scores)
        if debug:
            I = scores.argsort()[::-1]
            print("Q  ", query)
            for i in I:
                seq = seqs_sel[i]
                res = results[i]
                print("*  ", seq, res)

        i = scores.argmax()
        ties = scores == scores[i]
        return names_sel[ties], seqs_sel[ties], scores[ties]


class TieBreaker:
    def __init__(self, fname, place="left"):
        self.logger = logging.getLogger("TieBreaker")
        self.matcher = BarcodeMatcher(fname, place=place)
        self.query_count = defaultdict(float)
        self.bc_count = defaultdict(float)
        self.cache = {}
        self.n_hit = 0
        self.n_align = 0

    def align(self, query, debug=False, w=1):
        self.query_count[query] += w
        if not query in self.cache:
            self.n_align += w

            names, seqs, scores = self.matcher.align(query, debug)
            if debug:
                for n, s, S in zip(names, seqs, scores):
                    print(f"{n}\t{s}\t{S}")

            if (len(names) == 1) or (len(set(names)) == 1):
                # unambiguous best hit
                result = names[0], seqs[0], scores[0]
            else:
                # potentially do more involved resolving?
                result = (NO_CALL, NO_CALL, scores[0])

            self.cache[query] = result

        else:
            self.n_hit += w
            result = self.cache[query]

        self.bc_count[result[0]] += w
        self.bc_count["total"] += w
        return result

    def align_choices(self, queries, debug=False):
        w = 1.0 / len(queries)
        results = [self.align(q, debug=debug, w=w) for q in queries]
        score = np.array([(r[0] != NO_CALL) * r[2] for r in results])
        i = score.argmax()
        ties = score == score.argmax()
        if ties.sum() > 1:
            return (NO_CALL, NO_CALL, -1), queries[0]
        else:
            return results[i], queries[i]

    def load_cache(self, fname):
        self.logger.debug(f"pre-populating alignment cache from '{fname}'")
        if fname:
            try:
                df = pd.read_csv(
                    fname,
                    sep="\t",
                    index_col=None,
                    names=["query", "seq", "name", "score", "count"],
                )
            except OSError as err:
                self.logger.warning(f"error while loading caches: {err}")
            else:
                for row in df.itertuples():
                    self.cache[row.query] = (row.name, row.seq, row.score)

        self.logger.debug(f"loaded {len(self.cache)} queries.")


def store_cache(fname, cache, query_count, mincount=2):
    with open(fname, "w") as f:
        print(cache.keys())
        for query in sorted(cache.keys()):
            if query_count[query] >= mincount:
                name, seq, score = cache[query]
                f.write(f"{query}\t{seq}\t{name}\t{score}\t{query_count[query]}\n")


def report_stats(N, prefix=""):
    for k, v in sorted(N.items()):
        logging.info(f"{prefix}{k}\t{v}\t{100.0 * v/N['total']:.2f}")


def match_BC1(bc1_matcher, seq, qstart, tstart, N, debug=False, threshold=0.5):
    if tstart == 0:  # the start of opseq primer is intact
        bc1 = seq[qstart - 8 : qstart]
        BC1, ref1, score1 = bc1_matcher.align(bc1, debug=debug)
    else:
        # we have a possible deletion, or mutation. Check both options
        bc1_choices = [
            seq[qstart - 8 : qstart],
            seq[qstart - tstart - 8 : qstart - tstart],
        ]  # deletion  # mutation
        (BC1, ref1, score1), bc1 = bc1_matcher.align_choices(bc1_choices)

    N[f"BC1_score_{score1}"] += 1
    if BC1 == NO_CALL:
        N["BC1_ambig"] += 1
    elif score1 < 2 * len(BC1) * threshold:
        N["BC1_low_score"] += 1
        BC1 = NO_CALL
    else:
        N["BC1_assigned"] += 1

    return bc1, BC1, ref1, score1


def match_BC2(bc2_matcher, seq, qend, tend, N, threshold=0.5):
    bc2_choices = [
        seq[qend:],  # assume local alignment catches entire opseq
        seq[qend - 1 :],  # assume the last opseq base is in fact BC
        # (This indeed happens quite a bit for BC2)
    ]
    # if the end-gap is in fact a mutation, BC would start later
    if tend:
        mut = seq[qend + tend :]
        if len(mut) >= 8:
            bc2_choices.append(mut)

    # print(f"bc2 choices {bc2_choices}")
    (BC2, ref2, score2), bc2 = bc2_matcher.align_choices(bc2_choices)

    N[f"BC2_score_{score2}"] += 1
    if BC2 == NO_CALL:
        N["BC2_ambig"] += 1
    elif score2 < 2 * len(BC2) * threshold:
        N["BC2_low_score"] += 1
        BC2 = NO_CALL
    else:
        N["BC2_assigned"] += 1

    return bc2, BC2, ref2, score2


def opseq_local_align(
    seq,
    opseq="GAATCACGATACGTACACCAGT",
    min_opseq_score=22,
    min_start=8,
    max_end=22 + 12,
    allow_start_gap=False,
    allow_end_gap=False,
):
    results = pairwise2.align.localmd(
        opseq, seq, 2, -1.5, -3, -1, -3, -1, one_alignment_only=True, score_only=False
    )
    tstart = 0
    tend = 0
    if len(results) < 1:
        res = None
    else:
        res = results[0]
        qstart = res.start
        qend = res.end
        L = len(res.seqB)
        # print(res)
        # print(
        #     f"flags {qstart < min_start} {qend > max_end} {res.score < min_opseq_score}"
        # )
        if (
            (min_start and (qstart < min_start))
            or (max_end and (qend > max_end))
            or (res.score < min_opseq_score)
        ):
            res = None
        else:
            # check for start/end gaps
            # backtrack from qstart in seqA until we hit '-'
            if not allow_start_gap:
                while res.seqA[qstart - tstart - 1] != "-":
                    if qstart - tstart - 1 <= 0:
                        tstart = -1
                        res = None
                        break
                    tstart += 1

            if res is not None and not allow_end_gap:
                # and scan forward from qend until we hit '-'
                while res.seqA[qend + tend] != "-":
                    if qend + tend >= len(res.seqA) - 8:
                        tend = -1
                        res = None
                        break
                    tend += 1
    # print("end results", res, tstart, tend)
    return res, tstart, tend


def process_ordered_results(res_queue, args, Qerr, abort_flag):
    with ExceptionLogging("collector", Qerr=Qerr, exc_flag=abort_flag) as el:
        import heapq
        import time

        heap = []
        n_chunk_needed = 0
        t0 = time.time()
        t1 = t0
        n_rec = 0

        logger = el.logger
        out = Output(args)

        for n_chunk, results in queue_iter(res_queue, abort_flag):
            heapq.heappush(heap, (n_chunk, results))

            # as long as the root of the heap is the next needed chunk
            # pass results on to storage
            while heap and (heap[0][0] == n_chunk_needed):
                n_chunk, results = heapq.heappop(heap)  # retrieves heap[0]
                for assigned, record in results:
                    # print("record in process_ordered_results", record)
                    out.write(assigned, record)
                    n_rec += 1

                n_chunk_needed += 1

            # debug output on average throughput
            t2 = time.time()
            if t2 - t1 > 30:
                dT = t2 - t0
                rate = n_rec / dT
                logger.info(
                    "processed {0} reads in {1:.0f} seconds (average {2:.0f} reads/second).".format(
                        n_rec, dT, rate
                    )
                )
                t1 = t2

        out.close()
        # by the time None pops from the queue, all chunks
        # should have been processed!
        if not abort_flag.value:
            assert len(heap) == 0
        else:
            logger.warning(
                f"{len(heap)} chunks remained on the heap due to missing data upon abort."
            )

        dT = time.time() - t0
        logger.info(
            "finished processing {0} reads in {1:.0f} seconds (average {2:.0f} reads/second)".format(
                n_rec, dT, n_rec / dT
            )
        )


def process_fastq(Qfq, args, Qerr, abort_flag):
    """
    reads from two fastq files, groups the input into chunks for
    faster parallel processing, and puts these on a mp.Queue()
    """
    with ExceptionLogging("dispatcher", Qerr=Qerr, exc_flag=abort_flag) as el:
        for chunk in chunkify(read_source(args)):
            logging.debug(f"placing {chunk[0]} {len(chunk[1])} in queue")
            if put_or_abort(Qfq, chunk, abort_flag):
                el.logger.warning("shutdown flag was raised!")
                break


def count_dict_sum(sources):
    dst = defaultdict(float)
    for src in sources:
        for k, v in src.items():
            dst[k] += v

    return dst


def dict_merge(sources):
    dst = {}
    for src in sources:
        dst.update(src)
    return dst


def process_combinatorial(Qfq, Qres, args, Qerr, abort_flag, stat_lists):
    with ExceptionLogging("worker", Qerr=Qerr, exc_flag=abort_flag) as el:
        el.logger.debug(
            f"process_combinatorial starting up with Qfq={Qfq}, Qres={Qres} and args={args}"
        )
        bc1_matcher = TieBreaker(args.bc1_ref, place="left")
        bc2_matcher = TieBreaker(args.bc2_ref, place="right")
        bc1_matcher.load_cache(args.bc1_cache)
        bc2_matcher.load_cache(args.bc2_cache)

        out = Output(args, open_files=False)
        N = defaultdict(int)
        for n_chunk, reads in queue_iter(Qfq, abort_flag):
            el.logger.debug(f"received chunk {n_chunk} of {len(reads)} reads")
            results = []
            for fqid, r1, fqid2, r2, qual2 in reads:
                N["total"] += 1
                out_d = dict(qname=fqid, r1=r1, r2=r2, r2_qual=qual2, r2_qname=fqid2)
                # fallback values for bc1/bc2 so that some BC diversity
                # is maintained for debugging purposes in case we can not
                # assign a decent & unambiguous match
                # lower case bc is for original, uncorrected sequence
                out_d["bc1"] = r1[:12]
                out_d["bc2"] = r2[-12:]
                # upper case BC is for assigned, corrected sequence
                out_d["BC1"] = args.na
                out_d["BC2"] = args.na

                # align opseq sequence to seq of read1
                aln = opseq_local_align(
                    r1.rstrip(),
                    opseq=args.opseq,
                    min_opseq_score=args.min_opseq_score,
                    allow_end_gap=True,  # TODO more permanent fix for this quick'n'dirty hack to get short illumina read to work
                )
                res, tstart, tend = aln
                # print("OPSEQ", res, tstart, tend)
                if res is None:
                    N["opseq_broken"] += 1
                    assigned = False
                else:
                    # identify barcodes
                    bc1, BC1, ref1, score1 = match_BC1(
                        bc1_matcher,
                        res.seqB,
                        res.start,
                        tstart,
                        N,
                        threshold=args.threshold,
                    )
                    # bc2, BC2, ref2, score2 = match_BC2(
                    #     bc2_matcher, res.seqB, res.end, tend, N, threshold=args.threshold
                    # )
                    bc2, BC2, ref2, score2 = "na", "NA", "na", -1

                    # slo = sQSeq.lower()
                    # sout = slo[:qstart] + sQSeq[qstart:qend] + slo[qend:]
                    # print(sout, qstart, qend, tstart, tend, bc1, bc2)
                    # print(f"bc1: {bc1} -> {ref1} -> {BC1} score={50.0*score1/len(bc1):.1f} %")
                    # print(f"bc2: {bc2} -> {ref2} -> {BC2} score={50.0*score2/len(bc2):.1f} %")

                    # best matching pieces of sequence
                    out_d["bc1"] = bc1
                    out_d["bc2"] = bc2
                    # best attempt at assignment
                    out_d["BC1"] = BC1
                    out_d["BC2"] = BC2

                    if BC1 != NO_CALL:  # and BC2 != NO_CALL:
                        N["called"] += 1
                        assigned = True
                    else:
                        assigned = False

                out_d["assigned"] = assigned
                rec = out.make_record(**out_d)
                results.append((assigned, rec))

            Qres.put((n_chunk, results))

        N["BC1_cache_hit"] = bc1_matcher.n_hit
        N["BC2_cache_hit"] = bc2_matcher.n_hit

        # return our counts and observations
        # via these list proxies
        el.logger.debug("synchronizing cache and counts")
        Ns, qcaches1, qcaches2, qcounts1, qcounts2, bccounts1, bccounts2 = stat_lists
        Ns.append(N)
        qcaches1.append(bc1_matcher.cache)
        qcaches2.append(bc2_matcher.cache)
        qcounts1.append(bc1_matcher.query_count)
        qcounts2.append(bc2_matcher.query_count)
        bccounts1.append(bc1_matcher.bc_count)
        bccounts2.append(bc2_matcher.bc_count)


def log_qerr(qerr):
    "helper function for reporting errors in sub processes"
    for name, lines in qerr:
        for line in lines:
            logging.error(f"subprocess {name} exception {line}")


def main_combinatorial(args):

    # queues for communication between processes
    Qfq = mp.Queue(
        args.parallel * 5
    )  # FASTQ reads from process_fastq->process_combinatorial
    Qres = mp.Queue()  # extracted BCs from process_combinatorial->collector
    Qerr = mp.Queue()  # child-processes can report errors back to the main process here

    # Proxy objects to allow workers to report statistics about the run
    manager = mp.Manager()
    abort_flag = mp.Value("b")
    abort_flag.value = False
    qcaches1 = manager.list()
    qcaches2 = manager.list()
    qcounts1 = manager.list()
    qcounts2 = manager.list()
    bccounts1 = manager.list()
    bccounts2 = manager.list()
    Ns = manager.list()
    stat_lists = [Ns, qcaches1, qcaches2, qcounts1, qcounts2, bccounts1, bccounts2]

    with ExceptionLogging("main_combinatorial", exc_flag=abort_flag) as el:

        # read FASTQ in chunks and put them in Qfq
        dispatcher = mp.Process(
            target=process_fastq, name="dispatcher", args=(Qfq, args, Qerr, abort_flag)
        )

        dispatcher.start()
        el.logger.info("Started dispatch")

        # workers consume chunks of FASTQ from Qfq,
        # process them, and put the results in Qres
        workers = []
        for i in range(args.parallel):
            w = mp.Process(
                target=process_combinatorial,
                name=f"worker_{i}",
                args=(Qfq, Qres, args, Qerr, abort_flag, stat_lists),
            )
            w.start()
            workers.append(w)

        el.logger.info("Started workers")
        collector = mp.Process(
            target=process_ordered_results,
            name="output",
            args=(Qres, args, Qerr, abort_flag),
        )
        collector.start()
        el.logger.info("Started collector")
        # wait until all sequences have been thrown onto Qfq
        qfq, qerr = join_with_empty_queues(dispatcher, [Qfq, Qerr], abort_flag)
        el.logger.info("The dispatcher exited")
        if qfq or qerr:
            el.logger.info(f"{len(qfq)} chunks were drained from Qfq upon abort.")
            log_qerr(qerr)

        # signal all workers to finish
        el.logger.info("Signalling all workers to finish")
        for n in range(args.parallel):
            Qfq.put(None)  # each worker consumes exactly one None

        for w in workers:
            # make sure all results are on Qres by waiting for
            # workers to exit. Or, empty queues if aborting.
            qres, qerr = join_with_empty_queues(w, [Qres, Qerr], abort_flag)
            if qres or qerr:
                el.logger.info(f"{len(qres)} chunks were drained from Qres upon abort.")
                log_qerr(qerr)

        el.logger.info(
            "All worker processes have joined. Signalling collector to finish."
        )
        # signal the collector to stop
        Qres.put(None)

        # and wait until all output has been generated
        collector.join()
        el.logger.info("Collector has joined. Merging worker statistics.")

        N = count_dict_sum(Ns)
        if N["total"]:
            el.logger.info(
                f"Run completed. Overall combinatorial barcode assignment "
                f"rate was {100.0 * N['called']/N['total']}"
            )
        else:
            el.logger.error("No reads were processed!")

        if args.update_cache:
            qcount1 = count_dict_sum(qcounts1)
            qcount2 = count_dict_sum(qcounts2)
            cache1 = dict_merge(qcaches1)
            cache2 = dict_merge(qcaches2)
            store_cache(args.bc1_cache, cache1, qcount1)
            store_cache(args.bc2_cache, cache2, qcount2)

        if args.save_stats:
            bccount1 = count_dict_sum(bccounts1)
            bccount2 = count_dict_sum(bccounts2)
            with open(args.save_stats, "w") as f:
                for k, v in sorted(N.items()):
                    f.write(f"freq\t{k}\t{v}\t{100.0 * v/max(N['total'], 1):.2f}\n")

                for k, v in sorted(bccount1.items()):
                    f.write(
                        f"BC1\t{k}\t{v}\t{100.0 * v/max(bccount1['total'], 1):.2f}\n"
                    )

                for k, v in sorted(bccount2.items()):
                    f.write(
                        f"BC2\t{k}\t{v}\t{100.0 * v/max(bccount2['total'], 1):.2f}\n"
                    )


def process_dropseq(Qfq, Qres, args, Qerr, abort_flag, stat_lists):
    with ExceptionLogging("worker", Qerr=Qerr, exc_flag=abort_flag) as el:
        el.logger.debug(
            f"process_dropseq starting up with Qfq={Qfq}, Qres={Qres} and args={args}"
        )
        out = Output(args, open_files=False)
        N = defaultdict(int)
        for n_chunk, reads in queue_iter(Qfq, abort_flag):
            el.logger.debug(f"received chunk {n_chunk} of {len(reads)} reads")
            results = []
            for fqid, r1, fqid2, r2, qual2 in reads:
                N["total"] += 1
                rec = out.make_record(
                    assigned=True,
                    qname=fqid,
                    r1=r1,
                    r2=r2,
                    r2_qual=qual2,
                    r2_qname=fqid2,
                )
                results.append((True, rec))

            Qres.put((n_chunk, results))

        # return our counts and observations
        # via these list proxies
        el.logger.debug("synchronizing cache and counts")
        Ns = stat_lists[0]
        Ns.append(N)


def main_dropseq(args):
    # queues for communication between processes
    Qfq = mp.Queue(
        args.parallel * 5
    )  # FASTQ reads from process_fastq->process_combinatorial
    Qres = mp.Queue()  # extracted BCs from process_combinatorial->collector
    Qerr = mp.Queue()  # child-processes can report errors back to the main process here

    # Proxy objects to allow workers to report statistics about the run
    manager = mp.Manager()
    abort_flag = mp.Value("b")
    abort_flag.value = False
    Ns = manager.list()
    stat_lists = [Ns]

    with ExceptionLogging("main_combinatorial", exc_flag=abort_flag) as el:

        # read FASTQ in chunks and put them in Qfq
        dispatcher = mp.Process(
            target=process_fastq, name="dispatcher", args=(Qfq, args, Qerr, abort_flag)
        )

        dispatcher.start()
        el.logger.info("Started dispatch")

        # workers consume chunks of FASTQ from Qfq,
        # process them, and put the results in Qres
        workers = []
        for i in range(args.parallel):
            w = mp.Process(
                target=process_dropseq,
                name=f"worker_{i}",
                args=(Qfq, Qres, args, Qerr, abort_flag, stat_lists),
            )
            w.start()
            workers.append(w)

        el.logger.info("Started workers")
        collector = mp.Process(
            target=process_ordered_results,
            name="output",
            args=(Qres, args, Qerr, abort_flag),
        )
        collector.start()
        el.logger.info("Started collector")
        # wait until all sequences have been thrown onto Qfq
        qfq, qerr = join_with_empty_queues(dispatcher, [Qfq, Qerr], abort_flag)
        el.logger.info("The dispatcher exited")
        if qfq or qerr:
            el.logger.info(f"{len(qfq)} chunks were drained from Qfq upon abort.")
            log_qerr(qerr)

        # signal all workers to finish
        el.logger.info("Signalling all workers to finish")
        for n in range(args.parallel):
            Qfq.put(None)  # each worker consumes exactly one None

        for w in workers:
            # make sure all results are on Qres by waiting for
            # workers to exit. Or, empty queues if aborting.
            qres, qerr = join_with_empty_queues(w, [Qres, Qerr], abort_flag)
            if qres or qerr:
                el.logger.info(f"{len(qres)} chunks were drained from Qres upon abort.")
                log_qerr(qerr)

        el.logger.info(
            "All worker processes have joined. Signalling collector to finish."
        )
        # signal the collector to stop
        Qres.put(None)

        # and wait until all output has been generated
        collector.join()
        el.logger.info("Collector has joined. Merging worker statistics.")

        N = count_dict_sum(Ns)
        if N["total"]:
            el.logger.info(f"Run completed, {N['total']} reads processed.")
        else:
            el.logger.error("No reads were processed!")

        # N = defaultdict(int)
        # out = Output(args)

        # # iterate query sequences
        # for n, (qname, r1, r2_qname, r2, r2_qual) in enumerate(read_source(args)):
        #     N["total"] += 1
        #     out.write(True, qname=qname, r1=r1, r2_qname=r2_qname, r2=r2, r2_qual=r2_qual)

        #     if n and n % 100000 == 0:
        #         report_stats(N)

        # out.close()

        # report_stats(N)
        if args.save_stats:
            with open(args.save_stats, "w") as f:
                for k, v in sorted(N.items()):
                    f.write(f"freq\t{k}\t{v}\t{100.0 * v/max(N['total'], 1):.2f}\n")

    return N


class Output:
    def __init__(self, args, open_files=True):
        assert Output.safety_check_eval(args.cell_raw)
        assert Output.safety_check_eval(args.cell)
        assert Output.safety_check_eval(args.UMI)
        self.cell_raw = args.cell_raw
        self.cell = args.cell
        self.UMI = args.UMI
        self.na = args.na

        # precompile functions for speed-up
        self.f_cell_raw = compile(self.cell_raw, "<string cell_raw>", "eval")
        self.f_cell = compile(self.cell, "<string cell>", "eval")
        self.f_UMI = compile(self.UMI, "<string UMI>", "eval")

        self.fq_qual = args.fq_qual
        self.bc_na = args.na

        self.tags = []
        for tag in args.bam_tags.split(","):
            self.tags.append(tag.split(":"))

        if args.out_format == "fastq":
            fopen = lambda x: open(x, "w")
            self._write_record = self.write_fastq
            self._make_record = self.make_fastq_record

        elif args.out_format == "bam":
            prog = os.path.basename(__file__)
            header = {
                "HD": {"VN": "1.6"},
                "PG": [
                    {
                        "ID": 0,
                        "PN": prog,
                        "CL": " ".join(sys.argv[1:]),
                        "VN": __version__,
                    },
                ],
                "RG": [
                    {"ID": "A", "SM": args.sample},
                    {"ID": "U", "SM": f"unassigned_{args.sample}"},
                ],
            }
            self.bam_header = pysam.AlignmentHeader.from_dict(header)
            fopen = lambda x: pysam.AlignmentFile(x, "wbu", header=header)
            self._write_record = self.write_bam
            self._make_record = self.make_bam_record
        else:
            raise ValueError(f"unsopported output format '{args.out_format}'")

        if open_files:
            self.out_assigned = fopen(args.out_assigned)
            if args.out_unassigned != args.out_assigned:
                self.out_unassigned = fopen(args.out_unassigned)
            else:
                self.out_unassigned = self.out_assigned

    @staticmethod
    def safety_check_eval(s, danger="();."):
        chars = set(list(s))
        if chars & set(list(danger)):
            return False
        else:
            return True

    def make_bam_record(self, **kw):
        # sys.stderr.write(f"r2_qual={r2_qual}\n")
        a = pysam.AlignedSegment(self.bam_header)
        # STAR does not like spaces in read names so we have to split
        a.query_name = kw["r2_qname"].split()[0]
        a.query_sequence = kw["r2"]
        a.flag = 4
        a.query_qualities = pysam.qualitystring_to_array(kw["r2_qual"])
        a.tags = [(name, templ.format(**kw)) for name, templ in self.tags]
        return a.to_string()

    def make_fastq_record(self, **kw):
        seq = kw["cell"] + kw["UMI"]
        qual = self.fq_qual * len(seq)
        return f"@{kw['qname']}\n{seq}\n+\n{qual}\n"

    def make_record(self, assigned=True, **kw):
        kw["raw"], kw["cell"], kw["UMI"] = self.format(**kw)
        kw["assigned"] = "A" if assigned else "U"
        return self._make_record(**kw)

    def write_bam(self, out, rec):
        out.write(pysam.AlignedSegment.fromstring(rec, self.bam_header))

    def write_fastq(self, out, rec):
        out.write(rec)

    def write(self, assigned, record):
        if assigned:
            self._write_record(self.out_assigned, record)
        else:
            self._write_record(self.out_unassigned, record)

    def format(
        self,
        qname="qname1",
        r2_qname="qname2",
        r2_qual="",
        bc1=None,
        bc2=None,
        BC1=None,
        BC2=None,
        r1="",
        r2="",
        **kw,
    ):
        if BC1 is None:
            BC1 = self.na

        if BC2 is None:
            BC2 = self.na

        # slightly concerned about security here...
        # at least all () and ; raise an assertion in __init__
        cell = eval(self.f_cell)
        raw = eval(self.f_cell_raw)
        UMI = eval(self.f_UMI)
        # cell = eval(self.cell)
        # raw = eval(self.cell_raw)
        # UMI = eval(self.UMI)

        if (cell is None) or (UMI is None):
            raise ValueError(
                f"one of cell=eval('{self.cell}')='{cell}' "
                f"UMI=eval('{self.UMI}')='{UMI}' evaluated to None"
            )

        return raw, cell, UMI

    def close(self):
        self.out_assigned.close()
        self.out_unassigned.close()


# class Process(mp.Process):
#     def __init__(self, *argc, **kw):
#         mp.Process.__init__(self, *argc, **kw)

#     def run(self):


def setup_logging(args):
    FORMAT = f"%(asctime)-20s\t%(levelname)s\t{args.sample}\t%(name)s\t%(message)s"
    # formatter = logging.Formatter(FORMAT)
    lvl = getattr(logging, args.log_level, logging.INFO)
    logging.basicConfig(level=lvl, format=FORMAT)

    fh = logging.FileHandler(filename=args.log_file, mode="a")
    fh.setFormatter(logging.Formatter(FORMAT))
    root = logging.getLogger("")
    root.addHandler(fh)

    logger = logging.getLogger("main")
    logger.info(f"starting raw read preprocessing run")
    for k, v in sorted(vars(args).items()):
        logger.info(f"cmdline arg\t{k}={v}")

    # TODO: store options in YAML for quick
    # re-run with same options
    # {sorted(vars(args).items())}


def parse_args():
    parser = argparse.ArgumentParser(
        description="Convert raw reads1 and reads2 FASTQ into a single BAM file with cell barcode and UMI as BAM-tags"
    )
    parser.add_argument(
        "--sample",
        default="NA",
        help="source from where to get read1 (FASTQ format)",
    )
    parser.add_argument(
        "--read1",
        default="/dev/stdin",
        help="source from where to get read1 (FASTQ format)",
    )
    parser.add_argument(
        "--read2", default="", help="source from where to get read2 (FASTQ format)"
    )
    parser.add_argument("--cell", default="r1[8:20][::-1]")
    parser.add_argument("--cell-raw", default="None")
    parser.add_argument("--UMI", default="r1[0:8]")
    parser.add_argument(
        "--out-format",
        default="bam",
        choices=["fastq", "bam"],
        help="'bam' for tagged, unaligned BAM, 'fastq' for preprocessed read1 as FASTQ",
    )
    parser.add_argument(
        "--out-assigned",
        default="/dev/stdout",
        help="output for successful assignments (default=/dev/stdout) ",
    )
    parser.add_argument(
        "--out-unassigned",
        default="/dev/stdout",
        help="output for un-successful assignments (default=/dev/stdout) ",
    )
    parser.add_argument(
        "--save-stats",
        default="preprocessing_stats.txt",
        help="store statistics in this file",
    )
    parser.add_argument(
        "--log-file",
        default="preprocessing_run.log",
        help="store all log messages in this file (default=None)",
    )
    parser.add_argument(
        "--log-level",
        default="INFO",
        help="set sensitivity of the python logging facility (default=INFO)",
    )
    parser.add_argument(
        "--na", default="NNNNNNNN", help="code for ambiguous or unaligned barcode"
    )
    parser.add_argument(
        "--fq-qual",
        default="E",
        help="phred qual for assigned barcode bases in FASTQ output (default='E')",
    )

    # combinatorial barcode specific options
    parser.add_argument(
        "--parallel", default=1, type=int, help="how many processes to spawn"
    )
    parser.add_argument(
        "--opseq",
        default="GAATCACGATACGTACACCAGT",
        help="opseq primer site sequence",
    )
    parser.add_argument(
        "--min-opseq-score",
        default=22,
        type=float,
        help="minimal score for opseq alignment (default 22 [half of max])",
    )
    parser.add_argument(
        "--threshold",
        default=0.5,
        type=float,
        help="score threshold for calling a match (rel. to max for a given length)",
    )
    parser.add_argument(
        "--bc1-ref", default="", help="load BC1 reference sequences from this FASTA"
    )
    parser.add_argument(
        "--bc2-ref", default="", help="load BC2 reference sequences from this FASTA"
    )
    parser.add_argument(
        "--bc1-cache", default="", help="load cached BC1 alignments from here"
    )
    parser.add_argument(
        "--bc2-cache", default="", help="load cached BC2 alignments from here"
    )
    parser.add_argument(
        "--update-cache",
        default=False,
        action="store_true",
        help="update cache when run is complete (default=False)",
    )
    # parser.add_argument("--template", default="{cell}{UMI}")

    # bam r2 tagging. SAM standard now supports CB=corrected cell barcode, CR=original cell barcode, and MI=molecule identifier/UMI
    parser.add_argument(
        "--bam-tags",
        default="CB:{cell},MI:{UMI},RG:{assigned}",
        help="raw, uncorrected cell barcode",
    )
    args = parser.parse_args()

    return args



import re
import sys
import os
import snakemake
import argparse
import yaml
import logging

from shutil import copyfile
from spacemake.project_df import ProjectDF, get_project_sample_parser
from spacemake.config import ConfigFile
from spacemake.util import message_aggregation
from spacemake.errors import SpacemakeError

config_path = "config.yaml"
project_df = "project_df.csv"

logger_name = "spacemake.main"
logger = logging.getLogger(logger_name)


def get_run_parser():
    parser = argparse.ArgumentParser(allow_abbrev=False, add_help=False)

    parser.add_argument(
        "--cores", default=1, type=int, help="number of cores to be used in total"
    )
    parser.add_argument(
        "--dryrun",
        "-n",
        action="store_true",
        help="invokes a dry snakemake run, printing only commands",
    )
    parser.add_argument(
        "--rerun-incomplete",
        "--ri",
        action="store_true",
        help="forces snakemake to rerun incompletely generated files",
    )
    parser.add_argument(
        "--keep-going",
        action="store_true",
        help="if a job fails, keep executing independent jobs",
    )
    parser.add_argument(
        "--printshellcmds",
        "-p",
        action="store_true",
        help="print shell commands for each rule, if exist",
    )
    parser.add_argument(
        "--touch",
        "-t",
        action="store_true",
        help="rather than running the rules, just touch each file",
    )
    parser.add_argument(
        "--with_fastqc",
        "-wfqc",
        action="store_true",
        help="Run also fastqc as part of the spacemake run",
    )

    return parser


def setup_init_parser(parent_parser):
    parser_init = parent_parser.add_parser(
        "init",
        help="initialise spacemake: create config files, download genomes and annotations",
    )
    parser_init.add_argument(
        "--root_dir",
        default="",
        help="where to output the results of the spacemake run. defaults to .",
    )
    parser_init.add_argument(
        "--temp_dir",
        default="/tmp",
        help="temporary directory used when running spacemake. defaults to /tmp",
    )
    parser_init.add_argument(
        "--download_species",
        default=False,
        help="if set, upon initialisation, spacemake will download the mouse and human genome and index",
        action="store_true",
    )
    parser_init.add_argument(
        "--dropseq_tools",
        help="absolute path to dropseq_tools directory",
        required=True,
    )
    parser_init.set_defaults(func=spacemake_init)

    return parser_init


def setup_run_parser(pdf, parent_parser):
    parser_run = parent_parser.add_parser(
        "run", help="run spacemake", parents=[get_run_parser()]
    )
    parser_run.set_defaults(func=lambda args: spacemake_run(pdf, args))

    # create subparsers
    parser_run_subparsers = parser_run.add_subparsers()

    downsampling_parser = parser_run_subparsers.add_parser(
        "downsample",
        help="run downsampling analysis for a list of projects/samples",
        parents=[get_project_sample_parser(allow_multiple=True), get_run_parser()],
    )
    downsampling_parser.set_defaults(downsample=True,
        func=lambda args: spacemake_run(pdf, args))

    parser_novosparc = parser_run_subparsers.add_parser(
        'novosparc',
        help='reconstruct the 2-d tissue with novosparc',
        parents=[
            get_run_parser(),
            get_project_sample_parser(
                    allow_multiple = False,
                    help_extra = ' of sample to reconstruct',
                 ),
        ],
    )
        
    parser_novosparc.add_argument(
        '--run_mode', type=str, help='the run_mode to be used for this sample ' +
        'for reconstruction. If left empty, the first run_mode for this ' +
        'sample will be used', required=False,
    )

    parser_novosparc.add_argument(
        '--umi_cutoff', type=str, help='umi_cutoff to be used for this sample ' +
        'for reconstruction. If left empty, the smallest umi_cutoff of a given ' +
        'run_mode will be used', required=False,
    )

    parser_novosparc.add_argument(
        '--reference_project_id', type=str, help='project_id of the reference atlas.' +
        'Has to be spatial, otherwise error will be raised', required=False,
    )

    parser_novosparc.add_argument(
        '--reference_sample_id', type=str, help='sample_id of the reference atlas.' +
        'Has to be spatial, otherwise error will be raised', required=False,
    )

    parser_novosparc.add_argument(
        '--reference_run_mode', type=str, help='the run_mode to be used for the ' +
        'reference sample. If empty, the first run_mode for the reference will be used',
        required=False,
    )

    parser_novosparc.add_argument(
        '--reference_umi_cutoff', type=str, help='the umi_cutoff to be used for this ' +
        'reference sample. If empty, the smallest umi_cutoff of the given run_mode ' +
        'will be used.', required=False,
    )

    parser_novosparc.set_defaults(novosparc_reconstruct=True,
        func=lambda args: spacemake_run(pdf, args))
    
    return parser_run

@message_aggregation(logger_name)
def spacemake_init(args):
    if os.path.isfile(config_path):
        msg = "spacemake has already been initiated in this directory.\n"
        msg += "use other commands to run and analyse samples."
        logger.info(msg)
        return 0
    initial_config = os.path.join(os.path.dirname(__file__), "data/config/config.yaml")

    # initialise config file
    cf = ConfigFile.from_yaml(initial_config)
    # update the file path
    cf.set_file_path(config_path)

    # add variables from args
    cf.variables["root_dir"] = args["root_dir"]
    cf.variables["temp_dir"] = args["temp_dir"]

    cf.variables['external_bin'] = {
        'dropseq_tools' : args['dropseq_tools']
    }

    cf.variables['microscopy_out'] = args.get('microscopy_out', '')

    if args["download_species"]:
        species_data_config_file = os.path.join(
            os.path.dirname(__file__), "data/config/species_data_url.yaml"
        )
        species_data_config = yaml.load(
            open(species_data_config_file), Loader=yaml.FullLoader
        )

        # add keys as species to config
        species = list(species_data_config.keys())
        snakemake_config = {"root_dir": ""}
        snakemake_config["species"] = species

        # define the pattern
        snakemake_config[
            "annotation_file_pattern"
        ] = "species_data/{species}/{species}_{data_type}.gtf"
        snakemake_config[
            "genome_file_pattern"
        ] = "species_data/{species}/{species}_{data_type}.fa"

        # the to be saved file paths
        species_info = {}
        species_files_exist = []

        for sp in species:
            if not sp in species_info.keys():
                species_info[sp] = {}
            for data_type in ["genome", "annotation"]:
                # save download url to be passed to snakemake
                snakemake_config[sp + "_" + data_type + "_url"] = species_data_config[
                    sp
                ][data_type]
                # save file path of the result
                species_file = snakemake_config[data_type + "_file_pattern"].format(
                    species=sp, data_type=data_type
                )
                species_info[sp][data_type] = species_file

                # add bool if file exists
                species_files_exist.append(os.path.isfile(species_file))

        # if any of the files are missing
        if not all(species_files_exist):
            # get the snakefile for downloading the .gtf and .fa files
            snakefile = os.path.join(
                os.path.dirname(__file__), "snakemake/species_init.smk"
            )
            # run snakemake: download species data and place them in the right place
            snakemake.snakemake(snakefile, cores=1, config=snakemake_config)

        for key, value in species_info.items():
            cf.add_variable(
                "species", key, genome=value["genome"], annotation=value["annotation"]
            )

    # copy visium_puck_barcode_file
    dest_visium_path = "puck_data/visium_barcode_positions.csv"
    logger.info(f"Moving visium barcodes to {dest_visium_path}")
    os.makedirs(os.path.dirname(dest_visium_path), exist_ok=True)
    copyfile(
        os.path.join(os.path.dirname(__file__), "data/visium_barcode_positions.csv"),
        dest_visium_path,
    )

    # save
    cf.dump()

def get_novosparc_variables(pdf, args):
    # assert that sample exists
    pdf.assert_sample(args['project_id'], args['sample_id'])

    def populate_variables_from_args(pdf, args, arg_prefix=''):
        # get sample info
        sample_info = pdf.get_sample_info(
            project_id=args[f'{arg_prefix}project_id'],
            sample_id=args[f'{arg_prefix}sample_id']
        )

        # populate return dictionary
        ret = {
            f'{arg_prefix}project_id': args[f'{arg_prefix}project_id'],
            f'{arg_prefix}sample_id': args[f'{arg_prefix}sample_id'],
        }

        # get run mode
        if f'{arg_prefix}run_mode' in args:
            ret[f'{arg_prefix}run_mode'] = args[f'{arg_prefix}run_mode']
        else:
            run_mode_name = sample_info['run_mode'][0]
            ret[f'{arg_prefix}run_mode'] = run_mode_name
            logger.info(f"No run_mode provided, using {run_mode_name}")

        run_mode = pdf.config.get_run_mode(ret[f'{arg_prefix}run_mode'])

        if f'{arg_prefix}umi_cutoff' not in args:
            umi_cutoff = run_mode.variables['umi_cutoff'][0]
            ret[f'{arg_prefix}umi_cutoff'] = umi_cutoff
            logger.info(f"No umi_cutoff provided, using {umi_cutoff}")
        else:
            ret[f'{arg_prefix}umi_cutoff'] = args[f'{arg_prefix}umi_cutoff']

        return ret
    
    ret = populate_variables_from_args(pdf, args)

    if 'reference_project_id' not in args or 'reference_sample_id' not in args:
        logger.info('No reference_project_id or reference_sample_id provided,' +
            ' running novosparc de-novo...')
        ret['reference_project_id'] = ''
        ret['reference_sample_id'] = ''
        ret['reference_umi_cutoff'] = ''
        ret['reference_run_mode'] = ''
    else:
        pdf.assert_sample(args['reference_project_id'],
            args['reference_sample_id'])

        logger.info("Using (project_id, sample_id)="+
            f"({args['reference_project_id']}, {args['reference_sample_id']})" +
            " reference, running novosparc with reference...")

        novosparc_ret = populate_variables_from_args(pdf, args,
            arg_prefix='reference_')

        ret = {**ret, **novosparc_ret}

    return ret

@message_aggregation(logger_name)
def spacemake_run(pdf, args):
    if not os.path.isfile(config_path):
        msg = "spacemake has not been initalised yet.\n"
        msg += "please run `spacemake init` to start a new project"

        raise SpacemakeError(msg)

    samples = []
    projects = []
    targets = ['all']
    with_fastqc = args.get("with_fastqc", False)

    downsample = args.get("downsample", False)
    novosparc_reconstruct = args.get('novosparc_reconstruct', False)

    if downsample:
        targets = ["downsample"]
        samples = args.get("sample_id_list", [])
        projects = args.get("project_id_list", [])

    if novosparc_reconstruct:
        targets = ["novosparc"]
        novosparc_variables = get_novosparc_variables(pdf, args)
    else:
        novosparc_variables = {}

    config_variables = {
        "project_df": pdf.file_path,
        "samples": samples,
        "projects": projects,
        "with_fastqc": with_fastqc,
    }

    # join config_variables and novosparc_variables
    # to flatten the directory
    config_variables = {**config_variables, **novosparc_variables}

    # get the snakefile
    snakefile = os.path.join(os.path.dirname(__file__), "snakemake/main.smk")
    # run snakemake
    run_successful = snakemake.snakemake(
        snakefile,
        configfiles=[config_path],
        cores=args["cores"],
        dryrun=args["dryrun"],
        targets=targets,
        touch=args["touch"],
        force_incomplete=args["rerun_incomplete"],
        keepgoing=args["keep_going"],
        printshellcmds=args["printshellcmds"],
        config=config_variables,
    )

    if run_successful is False:
        raise SpacemakeError("an error occurred while snakemake() ran")


#################
# DEFINE PARSER #
#################
# we add main parser to a dictionary
# so that we can later call the help function
# based on the sub-command. this is to print the
# -h (help) functionality if no parameters are provided,
# rather than printing an error.

main_parser = argparse.ArgumentParser(
    allow_abbrev=False,
    description="spacemake: bioinformatic pipeline for processing and analysis of spatial-transcriptomics data",
)

subparsers = main_parser.add_subparsers(help="sub-command help", dest="subcommand")

parser_run = None
parser_projects = None
parser_config = None
parser_init = None

##################
# SPACEMAKE INIT #
##################
parser_init = setup_init_parser(subparsers)

####################
# SPACEMAKE CONFIG #
####################
## spacemake_config args
from spacemake.config import setup_config_parser

if os.path.isfile(config_path):
    cf = ConfigFile.from_yaml(config_path)
    # save config file
    parser_config = setup_config_parser(cf, subparsers)

############################
# SPACEMAKE PROJECT/SAMPLE #
############################
from spacemake.project_df import setup_project_parser

if os.path.isfile(config_path):
    pdf = ProjectDF(project_df, cf)
    parser_projects = setup_project_parser(pdf, subparsers)

    #################
    # SPACEMAKE RUN #
    #################
    parser_run = setup_run_parser(pdf, subparsers)

def cmdline():
    args = main_parser.parse_args()

    parser_dict = {
        "init": parser_init,
        "config": parser_config,
        "projects": parser_projects,
        "run": parser_run,
        "main": main_parser,
    }

    # get the function to be run
    if "func" in args:
        func = args.func
    # else print help
    else:
        if args.subcommand is not None:
            parser_dict[args.subcommand].print_help()
        else:
            parser_dict["main"].print_help()
        return 0

    # get the args and delete the func key, get only set values
    args = {key: value for key, value in vars(args).items() if value is not None}
    args.pop("func", None)
    # pop also main,
    args.pop("subcommand", None)

    func(args)


if __name__ == "__main__":
    cmdline()



from anndata import AnnData
from novosparc.common import Tissue
from spacemake.preprocess import calculate_adata_metrics

def compute_neighbors(adata, min_dist=None, max_dist=None):
    '''Compute all direct neighbors of all spots in the adata. Currently
        tailored for 10X Visium.
    Args:
        adata: an AnnData object
        min_dist: int, minimum distance to consider neighbors
        max_dist: int, maximum distance to consider neighbors
    Returns:
        neighbors: a dictionary holding the spot IDs for every spot
    '''
    import numpy as np
    from scipy.spatial.distance import cdist

    # Calculate all Euclidean distances on the adata
    dist_mtrx = cdist(adata.obsm['spatial'],
                      adata.obsm['spatial'])
    neighbors = dict()

    for i in range(adata.obs.shape[0]):
        neighbors[i] = np.where((dist_mtrx[i,:] > min_dist) & (dist_mtrx[i,:] < max_dist))[0]

    return neighbors

def compute_islands(adata, min_umi):
    '''Find contiguity islands

    Args:
        adata: an AnnData object
        cluster: a cell type label to compute the islands for
    Returns:
        islands: A list of lists of spots forming contiguity islands
    '''
    import numpy as np
    import itertools

    # this is hard coded for now for visium, to have 6 neighbors per spot
    # TODO: define an iterative approach where the key is to have around 6 
    # neighbors per spot on average
    neighbors = compute_neighbors(adata, min_dist = 0, max_dist=3)
    spots_cluster = np.where(np.array(adata.obs['total_counts']) < min_umi)[0]

    # create a new dict holding only neighbors of the cluster
    islands = []

    for spot in neighbors:
        if spot not in spots_cluster:
            continue
        islands.append({spot}.union({x for x in neighbors[spot] if x in spots_cluster}))

    # merge islands with common spots
    island_spots = set(itertools.chain.from_iterable(islands)) 

    for each in island_spots:
        components = [x for x in islands if each in x]
        for i in components:
            islands.remove(i)
        islands += [list(set(itertools.chain.from_iterable(components)))]

    return islands

def detect_tissue(adata, min_umi):
    ''' Detect tissue: first find beads with at least min_umi UMIs, then detect island in the rest
    
    Args
        adata: an AnnData object, with spatial coordinates
        min_umi: integer, the min umi to be assigned as tissue bead by default
    Returns:
        tissue_indices: a list of indices which should be kept for this AnnData object
    '''
    import numpy as np

    islands = compute_islands(adata, min_umi)

    # find the sizes of the islands. remove the biggest, as either the tissue has a big hole in it
    # or there are not so many big islands in which case removal is OK.
    # to be evaluated later...
    island_sizes = [len(island) for island in islands]

    tissue_islands = np.delete(islands, np.argmax(island_sizes))

    # get the indices of the islands
    tissue_indices = np.where(np.array(adata.obs['total_counts']) >= min_umi)[0]

    tissue_indices = np.append(tissue_indices, np.hstack(tissue_islands))

    adata = adata[tissue_indices, :]

    return adata

def create_mesh(
    width,
    height,
    diameter,
    distance,
    push_x = 0,
    push_y = 0):
    import numpy as np

    distance_y = np.sqrt(3) * distance
    
    x_coord = np.arange(push_x, width + diameter, distance)
    y_coord = np.arange(push_y, height + diameter, distance_y)
    
    X, Y = np.meshgrid(x_coord, y_coord)
    xy = np.vstack((X.flatten(), Y.flatten())).T
    
    return xy

def create_meshed_adata(adata,
        width_um,
        spot_diameter_um = 55,
        spot_distance_um = 100,
        bead_diameter_um = 10,
        mesh_type = 'circle'
    ):
    import pandas as pd
    import scanpy as sc
    import numpy as np
    import anndata

    from sklearn.metrics.pairwise import euclidean_distances
    from scipy.sparse import csr_matrix, csc_matrix, vstack

    if not mesh_type in ['circle', 'hexagon']:
        raise ValueError(f'unrecognised mesh type {mesh_type}')

    if mesh_type == 'hexagon': 
        spot_diameter_um = np.sqrt(3) * spot_diameter_um
        spot_distance_um = spot_diameter_um

    coords = adata.obsm['spatial']

    top_left_corner = np.min(coords, axis=0)
    bottom_right_corner = np.max(coords, axis=0)
    width_px = abs(top_left_corner[0] - bottom_right_corner[0])
    height_px = abs(top_left_corner[1] - bottom_right_corner[1])

    # capture area width
    spot_radius_um = spot_diameter_um / 2

    um_by_px = width_um/width_px
    spot_diameter_px = spot_diameter_um / um_by_px
    spot_radius_px = spot_radius_um / um_by_px
    spot_distance_px = spot_distance_um / um_by_px

    height_um = height_px * um_by_px

    # create meshgrid with one radius push
    xy = create_mesh(width_um,
        height_um,
        spot_diameter_um,
        spot_distance_um,
        push_x = spot_radius_um,
        push_y = spot_radius_um)
    # create pushed meshgrid, pushed by 
    xy_pushed = create_mesh(width_um,
        height_um,
        spot_diameter_um,
        spot_distance_um,
        push_x = spot_radius_um + spot_distance_um / 2,
        push_y = spot_radius_um + np.sqrt(3) * spot_distance_um / 2)

    mesh = np.vstack((xy, xy_pushed))
    mesh_px = mesh/um_by_px
    # add the top_left_corner
    mesh_px = mesh_px + top_left_corner

    # example: the diameter of one visium spot is 55um. if we take any bead, which center
    # falls into that, assuming that a bead is 10um, we would in fact take all beads
    # within 65um diameter. for this reason, the max distance should be 45um/2 between
    # visium spot center and other beads
    max_distance_px = (spot_diameter_um - bead_diameter_um)/um_by_px/2

    distance_M = euclidean_distances(mesh_px, coords)

    if mesh_type == 'circle':
        # circle mesh type: we create spots in a hexagonal mesh
        # example: the diameter of one visium spot is 55um. if we take any bead, which center
        # falls into that, assuming that a bead is 10um, we would in fact take all beads
        # within 65um diameter. for this reason, the max distance should be 45um/2 between
        # visium spot center and other beads
        max_distance_px = (spot_diameter_um - bead_diameter_um)/um_by_px/2
        # new_ilocs contains the indices of the columns of the sparse matrix to be created
        # original_ilocs contains the column location of the original adata.X (csr_matrix)
        new_ilocs, original_ilocs = np.nonzero(distance_M < max_distance_px)
    elif mesh_type == 'hexagon':
        # we simply create a hex mesh, without holes. for each spot we find the hexagon
        # it belings to
        new_ilocs = np.argmin(distance_M, axis=0)
        original_ilocs = np.arange(new_ilocs.shape[0])
        # we need to sort the new ilocs so that they are in increasing order
        sorted_ix = np.argsort(new_ilocs)
        new_ilocs = new_ilocs[sorted_ix]
        original_ilocs = original_ilocs[sorted_ix]

    joined_C = adata.X[original_ilocs]

    # at which indices does the index in the newly created matrix change
    change_ix = np.where(new_ilocs[:-1] != new_ilocs[1:])[0] + 1

    #array of indices, split by which row they should go together

    ix_array = np.asarray(np.split(np.arange(new_ilocs.shape[0]), change_ix, axis=0), dtype='object')

    joined_C_sumed = vstack([csr_matrix(joined_C[ix_array[n].astype(int), :].sum(0)) for n in range(len(ix_array))])

    joined_coordinates = mesh_px[np.unique(new_ilocs)]

    adata_out = anndata.AnnData(csc_matrix(joined_C_sumed), 
        obs = pd.DataFrame({'x_pos': joined_coordinates[:, 0],
                            'y_pos': joined_coordinates[:, 1]}),
        var = adata.var)

    adata_out.obsm['spatial'] = joined_coordinates
    
    # rename index
    adata_out.obs.index.name = 'cell_bc'

    def summarise_adata_obs_column(adata, column, summary_fun=sum):
        vals_to_join = adata.obs[column].to_numpy()[original_ilocs]
        vals_joined = np.array(
            [summary_fun(vals_to_join[ix_array[n].astype(int)])
                for n in range(len(ix_array))])
        return vals_joined
    print(adata)

    # summarise and attach n_reads, calculate metrics (incl. pcr)
    calculate_adata_metrics(adata_out,
        # provide the n_reads as a parameter
        n_reads = summarise_adata_obs_column(adata, 'n_reads'))

    adata_out.obs['n_joined'] = [len(x) for x in ix_array]

    from statistics import mean

    for column in ['exact_entropy', 'theoretical_entropy', 'exact_compression',\
        'theoretical_compression']:
        adata_out.obs[column] = summarise_adata_obs_column(adata, column, mean)

    return adata_out




import numpy as np
import pandas as pd
import scanpy as sc
import squidpy as sq

from spacemake.spatial import detect_tissue

# expect fitlered .h5ad dge, with spatial coords attached, tissue detected
adata = sc.read_h5ad(snakemake.input[0])
umi_cutoff = int(snakemake.wildcards['umi_cutoff'])

# filter_umi or detect tissue
# if data spatial and detect_tissue=True
if 'spatial' in adata.obsm.keys() and snakemake.params['run_mode_variables']['detect_tissue']:
    adata = detect_tissue(adata, umi_cutoff)
    print('tissue detection')
else:
    print(f'filtering by umi cutoff: {umi_cutoff}')
    adata = adata[adata.obs.total_counts > umi_cutoff, :]

# make the var indices (gene names) and obs indices (cell barcode) unique
adata.obs_names_make_unique()
adata.var_names_make_unique()

# save the raw counts
adata.raw = adata

# identify highly variable genes if we have any observations
nrow, ncol = adata.shape

# require at least 1000 genes expressed in the sample and at least 100 cells
if nrow > 100 and ncol >= 1000:
    print('starting analysis')
    try:
        sc.pp.highly_variable_genes(adata, flavor='seurat_v3', n_top_genes=2000)
    except ValueError:
        sc.pp.highly_variable_genes(adata, flavor='seurat_v3', n_top_genes=1000, span = 1)
    
    # calculate log(cpm)
    print('normalising and log-scaling')
    sc.pp.normalize_total(adata, target_sum=1e4)
    sc.pp.log1p(adata, base=2)
    
    # PCA ANALYSIS
    print('calculating pca components')
    sc.tl.pca(adata, svd_solver='arpack')
    
    # get number of pcs-s identified. Sometimes can be smaller than 50, if
    # less than 50 cells pass the threshold
    n_pcs = adata.uns['pca']['variance'].size
    # limit the number of pcs to 50
    n_pcs = n_pcs if n_pcs < 40 else 40
    
    # Compute the neighborhood graph
    print('computing neighborhood graph')
    sc.pp.neighbors(adata, n_pcs=n_pcs)
    
    # compute UMAP
    # for a very low number of cells, scanpy will throw an error here
    try:
        print('dimensionality reduction')
        sc.tl.umap(adata)   
    except TypeError:
        pass

    # find out the clusters
    # restrict to max 20 clusters
    resolution = [0.4, 0.6, 0.8, 1.0, 1.2]

    print('clustering')


    if snakemake.params['is_spatial']:
        sq.gr.spatial_neighbors(adata, coord_type="generic")

    for res in resolution:
        try:
            res_key = 'leiden_' + str(res)
            
            sc.tl.leiden(adata, resolution = res, key_added = res_key)
            
            # finding marker genes
            print(f'ranking genes for resolution {res}')
            sc.tl.rank_genes_groups(adata, res_key, method='t-test', key_added = 'rank_genes_groups_' + res_key, pts=True,
                use_raw = False)
            if snakemake.params['is_spatial']:
                # calculate nhood enrichment from squidpy
                sq.gr.nhood_enrichment(adata, cluster_key=res_key)
        except ZeroDivisionError as e:
            pass



adata.write(snakemake.output[0])



import numpy as np
import pandas as pd
import scanpy as sc

# save expression values as a long_df
def create_long_df(expr_matrix, id_vars = ['cell_bc']):
    long_df = expr_matrix.melt(id_vars = id_vars, var_name = 'gene', value_name = 'expr')
    long_df = long_df[long_df.expr > 0]
    return long_df

##############
# LOAD ADATA #
##############

adata = sc.read(snakemake.input[0])

uns_keys = ['hvg', 'leiden', 'log1p', 'neighbors', 'pca', 'umap']

# all the keys have to be in adata.uns
adata_complete = any([key in adata.uns.keys() for key in uns_keys])
#################
# TOP10 markers #
#################
if not adata_complete:
    pd.DataFrame().to_csv(snakemake.output['cluster_markers'])
    pd.DataFrame().to_csv(snakemake.output['nhood_enrichment'])
else:
    res_keys = adata.obs.columns[adata.obs.columns.str.startswith('leiden_')]
    
    top_10_marker_dfs = []
    nhood_enrichment_dfs = []
    
    for res_key in res_keys:
        rank_key = 'rank_genes_groups_' + res_key
        
        df = pd.DataFrame(adata.uns[rank_key]['names'])\
                .melt(var_name = 'cluster', value_name = 'gene')
        
        for key in ['logfoldchanges', 'pvals', 'pvals_adj']:
            df_key = pd.DataFrame(adata.uns[rank_key][key])\
                .melt(var_name = 'cluster', value_name = key)
            df[key] = df_key[key]
            # set the index to gene-cluster pair
            
        df.set_index(['gene', 'cluster'], inplace=True)
            
        for key in ['pts', 'pts_rest']:
            # get the percentage expressed in cluster and rest
            df2 = adata.uns[rank_key][key]
            df2['gene'] = df2.index
            df2 = df2.melt(var_name='cluster', id_vars='gene')\
                .set_index(['gene', 'cluster'])
            
            df[key] = df2.loc[df.index].value
             
        df['resolution'] = res_key.split('_')[1]
        df.reset_index(inplace=True)
         
        top_10_marker_dfs.append(df)

        if snakemake.params['is_spatial']:
        # get nhood data
            df = pd.DataFrame(adata.uns[f'{res_key}_nhood_enrichment']['zscore'])
            df = pd.melt(df.reset_index(), id_vars='index')\
                .rename(columns={'index': 'cluster_a',
                                 'variable': 'cluster_b',
                                 'value': 'zscore'})
            df['resolution'] = res_key.split('_')[1]

            nhood_enrichment_dfs.append(df)
        
    pd.concat(top_10_marker_dfs).to_csv(snakemake.output['cluster_markers'], index=False)

    if snakemake.params['is_spatial']:
        pd.concat(nhood_enrichment_dfs).to_csv(snakemake.output['nhood_enrichment'], index=False)
    else:
        # output empty csv file
        pd.DataFrame().to_csv(snakemake.output['nhood_enrichment'])


###############
# SAVE OBS DF #
###############
obs_df = adata.obs

if adata_complete:
    obs_df = sc.get.obs_df(adata, obsm_keys=[('X_umap', 0), ('X_umap', 1)])\
        .join(obs_df)\
        .rename(columns={'X_umap-0':'umap_0', 'X_umap-1':'umap_1'})

obs_df.index.set_names('cell_bc', inplace=True)

obs_df.to_csv(snakemake.output['obs_df'])

###############
# SAVE VAR DF #
###############
adata.var.index.set_names('gene_name', inplace=True)

adata.var.to_csv(snakemake.output['var_df'])
